#
# Copyright 2015-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Automatic identification of lines and wavelength calibration"""

import itertools
import numpy as np
from numpy.polynomial import Polynomial
from scipy.special import comb

from ..display.iofunctions import readi
from ..display.iofunctions import readf
from ..display.matplotlib_qt import set_window_geometry
from ..display.pause_debugplot import pause_debugplot
from ..display.polfit_residuals import polfit_residuals_with_cook_rejection
from ..display.polfit_residuals import polfit_residuals_with_sigma_rejection
from ..robustfit import fit_theil_sen
from ..stats import summary
from ..stats import robust_std
from .solutionarc import CrLinear, WavecalFeature, SolutionArcCalibration
from .peaks_spectrum import find_peaks_spectrum
from .peaks_spectrum import refine_peaks_spectrum

xmin_previous = None
xmax_previous = None
ymin_previous = None
ymax_previous = None


def select_data_for_fit(list_of_wvfeatures):
    """Select information from valid arc lines to facilitate posterior fits.

    Parameters
    ----------
    list_of_wvfeatures : list (of WavecalFeature instances)
        A list of size equal to the number of identified lines, which
        elements are instances of the class WavecalFeature, containing
        all the relevant information concerning the line
        identification.

    Returns
    -------
    nfit : int
        Number of valid points for posterior fits.
    ifit : list of int
        List of indices corresponding to the arc lines which
        coordinates are going to be employed in the posterior fits.
    xfit : 1d numpy aray
        X coordinate of points for posterior fits.
    yfit : 1d numpy array
        Y coordinate of points for posterior fits.
    wfit : 1d numpy array
        Cost function of points for posterior fits. The inverse of
        these values can be employed for weighted fits.

    """

    nlines_arc = len(list_of_wvfeatures)

    nfit = 0
    ifit = []
    xfit = np.array([])
    yfit = np.array([])
    wfit = np.array([])
    for i in range(nlines_arc):
        if list_of_wvfeatures[i].line_ok:
            ifit.append(i)
            xfit = np.append(xfit, [list_of_wvfeatures[i].xpos])
            yfit = np.append(yfit, [list_of_wvfeatures[i].reference])
            wfit = np.append(wfit, [list_of_wvfeatures[i].funcost])
            nfit += 1

    return nfit, ifit, xfit, yfit, wfit


def fit_list_of_wvfeatures(list_of_wvfeatures,
                           naxis1_arc,
                           crpix1,
                           poly_degree_wfit,
                           weighted=False,
                           plot_title=None,
                           geometry=None,
                           debugplot=0):
    """Fit polynomial to arc calibration list_of_wvfeatures.

    Parameters
    ----------
    list_of_wvfeatures : list (of WavecalFeature instances)
        A list of size equal to the number of identified lines, which
        elements are instances of the class WavecalFeature, containing
        all the relevant information concerning the line
        identification.
    naxis1_arc : int
        NAXIS1 of arc spectrum.
    crpix1 : float
        CRPIX1 value to be employed in the wavelength calibration.
    poly_degree_wfit : int
        Polynomial degree corresponding to the wavelength calibration
        function to be fitted.
    weighted : bool
        Determines whether the polynomial fit is weighted or not,
        using as weights the values of the cost function obtained in
        the line identification. Since the weights can be very
        different, typically weighted fits are not good because, in
        practice, they totally ignore the points with the smallest
        weights (which, in the other hand, are useful when handling
        the borders of the wavelength calibration range).
    plot_title : string or None
        Title for residuals plot.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    solution_wv : SolutionArcCalibration instance
        Instance of class SolutionArcCalibration, containing the
        information concerning the arc lines that have been properly
        identified. The information about all the lines (including
        those initially found but at the end discarded) is stored in
        the list of WavecalFeature instances 'list_of_wvfeatures'.

    """

    nlines_arc = len(list_of_wvfeatures)

    # select information from valid lines.
    nfit, ifit, xfit, yfit, wfit = select_data_for_fit(list_of_wvfeatures)

    # select list of filtered out and unidentified lines
    list_r = []
    list_t = []
    list_p = []
    list_k = []
    list_unidentified = []
    for i in range(nlines_arc):
        if not list_of_wvfeatures[i].line_ok:
            if list_of_wvfeatures[i].category == 'X':
                list_unidentified.append(i)
            elif list_of_wvfeatures[i].category == 'R':
                list_r.append(i)
            elif list_of_wvfeatures[i].category == 'T':
                list_t.append(i)
            elif list_of_wvfeatures[i].category == 'P':
                list_p.append(i)
            elif list_of_wvfeatures[i].category == 'K':
                list_k.append(i)
            else:
                raise ValueError('Unexpected "category"')

    # polynomial fit
    if weighted:
        weights = 1.0 / wfit
    else:
        weights = np.zeros_like(wfit) + 1.0

    if xfit.size <= poly_degree_wfit:
        raise ValueError("Insufficient number of points for fit.")
    poly, stats_list = Polynomial.fit(
        x=xfit, y=yfit, deg=poly_degree_wfit, full=True, w=weights
    )
    poly = Polynomial.cast(poly)
    coeff = poly.coef
    if len(xfit) > poly_degree_wfit + 1:
        residual_std = np.sqrt(stats_list[0]/(len(xfit)-poly_degree_wfit-1))[0]
    else:
        residual_std = 0.0

    if abs(debugplot) >= 10:
        print('>>> Fitted coefficients:\n', coeff)
        print('>>> Residual std.......:', residual_std)

    # obtain CRVAL1 and CDELT1 for a linear wavelength scale from the
    # last polynomial fit
    crval1_linear = poly(crpix1)
    crmin1_linear = poly(1)
    crmax1_linear = poly(naxis1_arc)
    cdelt1_linear = (crmax1_linear - crval1_linear) / (naxis1_arc - crpix1)
    if abs(debugplot) >= 10:
        print('>>> CRVAL1 linear scale:', crval1_linear)
        print('>>> CDELT1 linear scale:', cdelt1_linear)

    # generate solution (note that the class SolutionArcCalibration
    # only sotres the information in list_of_wvfeatures corresponding
    # to lines that have been properly identified
    cr_linear = CrLinear(
        crpix1,
        crval1_linear,
        crmin1_linear,
        crmax1_linear,
        cdelt1_linear
    )

    solution_wv = SolutionArcCalibration(
        features=list_of_wvfeatures,
        coeff=coeff,
        residual_std=residual_std,
        cr_linear=cr_linear
    )

    if abs(debugplot) % 10 != 0:
        # polynomial fit
        xpol = np.linspace(1, naxis1_arc, naxis1_arc)
        ypol = poly(xpol) - (crval1_linear + (xpol - crpix1) * cdelt1_linear)
        # identified lines
        xp = np.copy(xfit)
        yp = yfit - (crval1_linear + (xp - crpix1) * cdelt1_linear)
        yres = yfit - poly(xp)  # residuals
        # include residuals plot with identified lines
        from numina.array.display.matplotlib_qt import plt
        fig = plt.figure()
        ax2 = fig.add_subplot(2, 1, 2)
        ax2.set_xlim(1 - 0.05 * naxis1_arc, naxis1_arc + 0.05 * naxis1_arc)
        ax2.set_xlabel('pixel position in arc spectrum [from 1 to NAXIS1]')
        ax2.set_ylabel('residuals (Angstrom)')
        ax2.plot(xp, yres, 'go')
        ax2.axhline(y=0.0, color="black", linestyle="dashed")
        # residuals with R, T, P and K lines
        for val in zip(["R", "T", "P", "K"],
                       [list_r, list_t, list_p, list_k],
                       ['red', 'blue', 'magenta', 'orange']):
            list_x = val[1]
            if len(list_x) > 0:
                xxp = np.array([])
                yyp = np.array([])
                for i in list_x:
                    xxp = np.append(xxp, [list_of_wvfeatures[i].xpos])
                    yyp = np.append(yyp, [list_of_wvfeatures[i].reference])
                yyres = yyp - poly(xxp)
                ax2.plot(xxp, yyres, marker='x', markersize=15, c=val[2],
                         linewidth=0)

        # plot with differences between linear fit and fitted
        # polynomial
        ax = fig.add_subplot(2, 1, 1, sharex=ax2)
        ax.set_xlim(1 - 0.05 * naxis1_arc, naxis1_arc + 0.05 * naxis1_arc)
        ax.set_ylabel('differences with\nlinear solution (Angstrom)')
        ax.plot(xp, yp, 'go', label="identified")
        for i in range(nfit):
            ax.text(xp[i], yp[i], list_of_wvfeatures[ifit[i]].category,
                    fontsize=15)
        # polynomial fit
        ax.plot(xpol, ypol, 'c-', label="fit")
        # unidentified lines
        if len(list_unidentified) > 0:
            ymin = np.concatenate((yp, ypol)).min()
            ymax = np.concatenate((yp, ypol)).max()
            for i in list_unidentified:
                xxp = np.array([list_of_wvfeatures[i].xpos,
                                list_of_wvfeatures[i].xpos])
                yyp = np.array([ymin, ymax])
                if i == list_unidentified[0]:
                    ax.plot(xxp, yyp, 'r--', label='unidentified')
                else:
                    ax.plot(xxp, yyp, 'r--')
        # R, T, P and K lines
        for val in zip(["R", "T", "P", "K"],
                       [list_r, list_t, list_p, list_k],
                       ['red', 'blue', 'magenta', 'orange']):
            list_x = val[1]
            if len(list_x) > 0:
                xxp = np.array([])
                yyp = np.array([])
                for i in list_x:
                    xxp = np.append(xxp, [list_of_wvfeatures[i].xpos])
                    yyp = np.append(yyp, [list_of_wvfeatures[i].reference])
                yyp -= crval1_linear + (xxp - crpix1) * cdelt1_linear
                ax.plot(xxp, yyp, marker='x', markersize=15, c=val[2],
                        linewidth=0, label='removed')
                for k in range(len(xxp)):
                    ax.text(xxp[k], yyp[k], val[0], fontsize=15)

        # legend
        ax.legend()

        # title
        if plot_title is None:
            plt.title("Wavelength calibration")
        else:
            plt.title(plot_title)

        # include important parameters in plot
        ax.text(0.50, 0.25, "poldeg: " + str(poly_degree_wfit) +
                ", nfit: " + str(len(xfit)),
                fontsize=12,
                transform=ax.transAxes,
                horizontalalignment="center",
                verticalalignment="bottom")
        ax.text(0.50, 0.15, "CRVAL1: " + str(round(crval1_linear, 4)),
                fontsize=12,
                transform=ax.transAxes,
                horizontalalignment="center",
                verticalalignment="bottom")
        ax.text(0.50, 0.05, "CDELT1: " + str(round(cdelt1_linear, 4)),
                fontsize=12,
                transform=ax.transAxes,
                horizontalalignment="center",
                verticalalignment="bottom")
        ax2.text(0.50, 0.05, "r.m.s.: " + str(round(residual_std, 4)),
                 fontsize=12,
                 transform=ax2.transAxes,
                 horizontalalignment="center",
                 verticalalignment="bottom")

        # set window geometry
        set_window_geometry(geometry)
        pause_debugplot(debugplot, pltshow=True, tight_layout=False)

    return solution_wv


def gen_triplets_master(wv_master, geometry=None, debugplot=0):
    """Compute information associated to triplets in master table.

    Determine all the possible triplets that can be generated from the
    array `wv_master`. In addition, the relative position of the
    central line of each triplet is also computed.

    Parameters
    ----------
    wv_master : 1d numpy array, float
        Array with wavelengths corresponding to the master table
        (Angstroms).
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    ntriplets_master : int
        Number of triplets built from master table.
    ratios_master_sorted : 1d numpy array, float
        Array with values of the relative position of the central line
        of each triplet, sorted in ascending order.
    triplets_master_sorted_list : list of tuples
        List with tuples of three numbers, corresponding to the three
        line indices in the master table. The list is sorted to be in
        correspondence with `ratios_master_sorted`.

    """

    nlines_master = wv_master.size

    # Check that the wavelengths in the master table are sorted
    wv_previous = wv_master[0]
    for i in range(1, nlines_master):
        if wv_previous >= wv_master[i]:
            raise ValueError('Wavelengths:\n--> ' +
                             str(wv_previous) + '\n--> ' + str(wv_master[i]) +
                             '\nin master table are duplicated or not sorted')
        wv_previous = wv_master[i]

    # Generate all the possible triplets with the numbers of the lines
    # in the master table. Each triplet is defined as a tuple of three
    # numbers corresponding to the three line indices in the master
    # table. The collection of tuples is stored in an ordinary python
    # list.
    iter_comb_triplets = itertools.combinations(range(nlines_master), 3)
    triplets_master_list = [val for val in iter_comb_triplets]

    # Verify that the number of triplets coincides with the expected
    # value.
    ntriplets_master = len(triplets_master_list)
    if ntriplets_master == comb(nlines_master, 3, exact=True):
        if abs(debugplot) >= 10:
            print('>>> Total number of lines in master table:',
                  nlines_master)
            print('>>> Number of triplets in master table...:',
                  ntriplets_master)
    else:
        raise ValueError('Invalid number of combinations')

    # For each triplet, compute the relative position of the central
    # line.
    ratios_master = np.zeros(ntriplets_master)
    for index, value in enumerate(triplets_master_list):
        i1, i2, i3 = value
        delta1 = wv_master[i2] - wv_master[i1]
        delta2 = wv_master[i3] - wv_master[i1]
        ratios_master[index] = delta1 / delta2

    # Compute the array of indices that index the above ratios in
    # sorted order.
    isort_ratios_master = np.argsort(ratios_master)

    # Simultaneous sort of position ratios and triplets.
    ratios_master_sorted = ratios_master[isort_ratios_master]
    triplets_master_sorted_list = [triplets_master_list[i]
                                   for i in isort_ratios_master]

    if abs(debugplot) in [21, 22]:
        # compute and plot histogram with position ratios
        bins_in = np.linspace(0.0, 1.0, 41)
        hist, bins_out = np.histogram(ratios_master, bins=bins_in)
        #
        from numina.array.display.matplotlib_qt import plt
        fig = plt.figure()
        ax = fig.add_subplot(111)
        width_hist = 0.8*(bins_out[1]-bins_out[0])
        center = (bins_out[:-1]+bins_out[1:])/2
        ax.bar(center, hist, align='center', width=width_hist)
        ax.set_xlabel('distance ratio in each triplet')
        ax.set_ylabel('Number of triplets')
        ax.set_title("Number of lines/triplets: " +
                     str(nlines_master) + "/" + str(ntriplets_master))
        # set window geometry
        set_window_geometry(geometry)
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

    return ntriplets_master, ratios_master_sorted, triplets_master_sorted_list


def arccalibration(wv_master,
                   xpos_arc,
                   naxis1_arc,
                   crpix1,
                   wv_ini_search,
                   wv_end_search,
                   wvmin_useful,
                   wvmax_useful,
                   error_xpos_arc,
                   times_sigma_r,
                   frac_triplets_for_sum,
                   times_sigma_theil_sen,
                   poly_degree_wfit,
                   times_sigma_polfilt,
                   times_sigma_cook,
                   times_sigma_inclusion,
                   geometry=None,
                   debugplot=0):
    """Performs arc line identification for arc calibration.

    This function is a wrapper of two functions, which are responsible
    of computing all the relevant information concerning the triplets
    generated from the master table and the actual identification
    procedure of the arc lines, respectively.

    The separation of those computations in two different functions
    helps to avoid the repetition of calls to the first function when
    calibrating several arcs using the same master table.

    Parameters
    ----------
    wv_master : 1d numpy array, float
        Array with wavelengths corresponding to the master table
        (Angstroms).
    xpos_arc : 1d numpy array, float
        Location of arc lines (pixels).
    naxis1_arc : int
        NAXIS1 for arc spectrum.
    crpix1 : float
        CRPIX1 value to be employed in the wavelength calibration.
    wv_ini_search : float
        Minimum expected wavelength in spectrum.
    wv_end_search : float
        Maximum expected wavelength in spectrum.
    wvmin_useful : float
        If not None, this value is used to clip detected lines below it.
    wvmax_useful : float
        If not None, this value is used to clip detected lines above it.
    error_xpos_arc : float
        Error in arc line position (pixels).
    times_sigma_r : float
        Times sigma to search for valid line position ratios.
    frac_triplets_for_sum : float
        Fraction of distances to different triplets to sum when
        computing the cost function.
    times_sigma_theil_sen : float
        Number of times the (robust) standard deviation around the
        linear fit (using the Theil-Sen method) to reject points.
    poly_degree_wfit : int
        Degree for polynomial fit to wavelength calibration.
    times_sigma_polfilt : float
        Number of times the (robust) standard deviation around the
        polynomial fit to reject points.
    times_sigma_cook : float
        Number of times the standard deviation of Cook's distances
        to detect outliers. If zero, this method of outlier detection
        is ignored.
    times_sigma_inclusion : float
        Number of times the (robust) standard deviation around the
        polynomial fit to include a new line in the set of identified
        lines.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    list_of_wvfeatures : list (of WavecalFeature instances)
        A list of size equal to the number of identified lines, which
        elements are instances of the class WavecalFeature, containing
        all the relevant information concerning the line
        identification.

    """

    ntriplets_master, ratios_master_sorted, triplets_master_sorted_list = \
        gen_triplets_master(wv_master=wv_master, geometry=geometry,
                            debugplot=debugplot)

    list_of_wvfeatures = arccalibration_direct(
        wv_master=wv_master,
        ntriplets_master=ntriplets_master,
        ratios_master_sorted=ratios_master_sorted,
        triplets_master_sorted_list=triplets_master_sorted_list,
        xpos_arc=xpos_arc,
        naxis1_arc=naxis1_arc,
        crpix1=crpix1,
        wv_ini_search=wv_ini_search,
        wv_end_search=wv_end_search,
        wvmin_useful=wvmin_useful,
        wvmax_useful=wvmax_useful,
        error_xpos_arc=error_xpos_arc,
        times_sigma_r=times_sigma_r,
        frac_triplets_for_sum=frac_triplets_for_sum,
        times_sigma_theil_sen=times_sigma_theil_sen,
        poly_degree_wfit=poly_degree_wfit,
        times_sigma_polfilt=times_sigma_polfilt,
        times_sigma_cook=times_sigma_cook,
        times_sigma_inclusion=times_sigma_inclusion,
        geometry=geometry,
        debugplot=debugplot)

    return list_of_wvfeatures


def arccalibration_direct(wv_master,
                          ntriplets_master,
                          ratios_master_sorted,
                          triplets_master_sorted_list,
                          xpos_arc,
                          naxis1_arc,
                          crpix1,
                          wv_ini_search,
                          wv_end_search,
                          wvmin_useful=None,
                          wvmax_useful=None,
                          error_xpos_arc=1.0,
                          times_sigma_r=3.0,
                          frac_triplets_for_sum=0.50,
                          times_sigma_theil_sen=10.0,
                          poly_degree_wfit=3,
                          times_sigma_polfilt=10.0,
                          times_sigma_cook=10.0,
                          times_sigma_inclusion=5.0,
                          geometry=None,
                          debugplot=0):
    """Performs line identification for arc calibration using line triplets.

    This function assumes that a previous call to the function
    responsible for the computation of information related to the
    triplets derived from the master table has been previously
    executed.

    Parameters
    ----------
    wv_master : 1d numpy array, float
        Array with wavelengths corresponding to the master table
        (Angstroms).
    ntriplets_master : int
        Number of triplets built from master table.
    ratios_master_sorted : 1d numpy array, float
        Array with values of the relative position of the central line
        of each triplet, sorted in ascending order.
    triplets_master_sorted_list : list of tuples
        List with tuples of three numbers, corresponding to the three
        line indices in the master table. The list is sorted to be in
        correspondence with `ratios_master_sorted`.
    xpos_arc : 1d numpy array, float
        Location of arc lines (pixels).
    naxis1_arc : int
        NAXIS1 for arc spectrum.
    crpix1 : float
        CRPIX1 value to be employed in the wavelength calibration.
    wv_ini_search : float
        Minimum expected wavelength in spectrum.
    wv_end_search : float
        Maximum expected wavelength in spectrum.
    wvmin_useful : float or None
        If not None, this value is used to clip detected lines below it.
    wvmax_useful : float or None
        If not None, this value is used to clip detected lines above it.
    error_xpos_arc : float
        Error in arc line position (pixels).
    times_sigma_r : float
        Times sigma to search for valid line position ratios.
    frac_triplets_for_sum : float
        Fraction of distances to different triplets to sum when
        computing the cost function.
    times_sigma_theil_sen : float
        Number of times the (robust) standard deviation around the
        linear fit (using the Theil-Sen method) to reject points.
    poly_degree_wfit : int
        Degree for polynomial fit to wavelength calibration.
    times_sigma_polfilt : float
        Number of times the (robust) standard deviation around the
        polynomial fit to reject points.
    times_sigma_cook : float
        Number of times the standard deviation of Cook's distances
        to detect outliers. If zero, this method of outlier detection
        is ignored.
    times_sigma_inclusion : float
        Number of times the (robust) standard deviation around the
        polynomial fit to include a new line in the set of identified
        lines.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    list_of_wvfeatures : list (of WavecalFeature instances)
        A list of size equal to the number of identified lines, which
        elements are instances of the class WavecalFeature, containing
        all the relevant information concerning the line
        identification.

    """

    nlines_master = wv_master.size

    delta_wv = 0.20 * (wv_master.max() - wv_master.min())
    if wv_ini_search is None:
        wv_ini_search = wv_master.min() - delta_wv
    if wv_end_search is None:
        wv_end_search = wv_master.max() + delta_wv

    nlines_arc = xpos_arc.size
    if nlines_arc < 5:
        raise ValueError('Insufficient arc lines=' + str(nlines_arc))

    # ---
    # Generate triplets with consecutive arc lines. For each triplet,
    # compatible triplets from the master table are sought. Each
    # compatible triplet from the master table provides an estimate for
    # CRVAL1 and CDELT1. As an additional constraint, the only valid
    # solutions are those for which the initial and the final
    # wavelengths for the arc are restricted to a predefined wavelength
    # interval.
    crval1_search = np.array([])
    cdelt1_search = np.array([])
    error_crval1_search = np.array([])
    error_cdelt1_search = np.array([])
    itriplet_search = np.array([], dtype=int)
    clabel_search = []

    ntriplets_arc = nlines_arc - 2
    if abs(debugplot) >= 10:
        print('>>> Total number of arc lines............:', nlines_arc)
        print('>>> Total number of arc triplets.........:', ntriplets_arc)

    # maximum allowed value for CDELT1
    cdelt1_max = (wv_end_search-wv_ini_search)/float(naxis1_arc-1)

    # Loop in all the arc line triplets. Note that only triplets built
    # from consecutive arc lines are considered.
    for i in range(ntriplets_arc):
        i1, i2, i3 = i, i+1, i+2

        dist12 = xpos_arc[i2] - xpos_arc[i1]
        dist13 = xpos_arc[i3] - xpos_arc[i1]
        ratio_arc = dist12 / dist13

        pol_r = ratio_arc * (ratio_arc - 1) + 1
        error_ratio_arc = np.sqrt(2) * error_xpos_arc/dist13 * np.sqrt(pol_r)

        ratio_arc_min = max(0.0, ratio_arc-times_sigma_r*error_ratio_arc)
        ratio_arc_max = min(1.0, ratio_arc+times_sigma_r*error_ratio_arc)

        # determine compatible triplets from the master list
        j_loc_min = np.searchsorted(ratios_master_sorted, ratio_arc_min)-1
        j_loc_max = np.searchsorted(ratios_master_sorted, ratio_arc_max)+1

        if j_loc_min < 0:
            j_loc_min = 0
        if j_loc_max > ntriplets_master:
            j_loc_max = ntriplets_master

        if abs(debugplot) >= 10:
            print(i, ratio_arc_min, ratio_arc, ratio_arc_max,
                  j_loc_min, j_loc_max)

        # each triplet from the master list provides a potential
        # solution for CRVAL1 and CDELT1
        for j_loc in range(j_loc_min, j_loc_max):
            j1, j2, j3 = triplets_master_sorted_list[j_loc]
            # initial solutions for CDELT1, CRVAL1 and CRMAX1
            cdelt1_temp = (wv_master[j3]-wv_master[j1])/dist13
            crval1_temp = wv_master[j2]-(xpos_arc[i2]-crpix1)*cdelt1_temp
            crmin1_temp = crval1_temp + float(1-crpix1)*cdelt1_temp
            # crmax1_temp = crval1_temp + float(naxis1_arc-crpix1)*cdelt1_temp
            # check that CRMIN1 and CRMAX1 are within the valid limits
            if wv_ini_search <= crmin1_temp <= wv_end_search \
                    and cdelt1_temp <= cdelt1_max:
                # Compute errors
                error_crval1_temp = \
                    cdelt1_temp*error_xpos_arc * \
                    np.sqrt(1+2*((xpos_arc[i2]-crpix1)**2)/(dist13**2))
                error_cdelt1_temp = \
                    np.sqrt(2)*cdelt1_temp * error_xpos_arc/dist13
                # Store values and errors
                crval1_search = np.append(crval1_search, [crval1_temp])
                cdelt1_search = np.append(cdelt1_search, [cdelt1_temp])
                error_crval1_search = np.append(error_crval1_search,
                                                [error_crval1_temp])
                error_cdelt1_search = np.append(error_cdelt1_search,
                                                [error_cdelt1_temp])
                # Store additional information about the triplets
                itriplet_search = np.append(itriplet_search, [i])
                clabel_search.append((j1, j2, j3))

    # normalize the values of CDELT1 and CRVAL1 to the interval [0,1]
    # in each case
    cdelt1_search_norm = cdelt1_search/cdelt1_max
    error_cdelt1_search_norm = error_cdelt1_search/cdelt1_max
    #
    crval1_search_norm = (crval1_search-wv_ini_search)
    crval1_search_norm /= (wv_end_search-wv_ini_search)
    error_crval1_search_norm = error_crval1_search
    error_crval1_search_norm /= (wv_end_search-wv_ini_search)

    # intermediate plots
    if abs(debugplot) in [21, 22]:
        from numina.array.display.matplotlib_qt import plt

        # CDELT1 vs CRVAL1 diagram (original coordinates)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('cdelt1 (Angstroms/pixel)')
        ax.set_ylabel('crval1 (Angstroms)')
        ax.scatter(cdelt1_search, crval1_search, s=200, alpha=0.1)
        xmin = 0.0
        xmax = cdelt1_max
        dx = xmax-xmin
        xmin -= dx/20
        xmax += dx/20
        ax.set_xlim(xmin, xmax)
        ymin = wv_ini_search
        ymax = wv_end_search
        dy = ymax-ymin
        ymin -= dy/20
        ymax += dy/20
        ax.set_ylim(ymin, ymax)
        xp_limits = np.array([0., cdelt1_max])
        yp_limits = wv_end_search-float(naxis1_arc-1)*xp_limits
        xp_limits = np.concatenate((xp_limits, [xp_limits[0], xp_limits[0]]))
        yp_limits = np.concatenate((yp_limits, [yp_limits[1], yp_limits[0]]))
        ax.plot(xp_limits, yp_limits, linestyle='-', color='magenta')
        ax.set_title("Potential solutions within the valid parameter space")
        # set window geometry
        set_window_geometry(geometry)
        print('Number of points in last plot:', len(cdelt1_search))
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

        # CDELT1 vs CRVAL1 diagram (normalized coordinates)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('normalized cdelt1')
        ax.set_ylabel('normalized crval1')
        ax.scatter(cdelt1_search_norm, crval1_search_norm, s=200, alpha=0.1)
        xmin = -0.05
        xmax = 1.05
        ymin = -0.05
        ymax = 1.05
        xp_limits = np.array([0., 1., 0., 0.])
        yp_limits = np.array([1., 0., 0., 1.])
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.plot(xp_limits, yp_limits, linestyle='-', color='magenta')
        ax.set_title("Potential solutions within the valid parameter space")
        # set window geometry
        set_window_geometry(geometry)
        print('Number of points in last plot:', len(cdelt1_search_norm))
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

        # CDELT1 vs CRVAL1 diagram (normalized coordinates)
        # with different color for each arc triplet and overplotting
        # the arc triplet number
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('normalized cdelt1')
        ax.set_ylabel('normalized crval1')
        ax.scatter(cdelt1_search_norm, crval1_search_norm, s=200, alpha=0.1,
                   c=itriplet_search)
        for i in range(len(itriplet_search)):
            ax.text(cdelt1_search_norm[i], crval1_search_norm[i],
                    str(int(itriplet_search[i])), fontsize=6)
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.plot(xp_limits, yp_limits, linestyle='-', color='magenta')
        ax.set_title("Potential solutions: arc line triplet number")
        # set window geometry
        set_window_geometry(geometry)
        print('Number of points in last plot:', len(cdelt1_search_norm))
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

        # CDELT1 vs CRVAL1 diagram (normalized coordinates)
        # including triplet numbers
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('normalized cdelt1')
        ax.set_ylabel('normalized crval1')
        ax.scatter(cdelt1_search_norm, crval1_search_norm, s=200, alpha=0.1,
                   c=itriplet_search)
        for i in range(len(clabel_search)):
            ax.text(cdelt1_search_norm[i], crval1_search_norm[i],
                    clabel_search[i], fontsize=6)
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.plot(xp_limits, yp_limits, linestyle='-', color='magenta')
        ax.set_title("Potential solutions: master line triplets")
        # set window geometry
        set_window_geometry(geometry)
        print('Number of points in last plot:', len(cdelt1_search_norm))
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

        # CDELT1 vs CRVAL1 diagram (normalized coordinates)
        # with error bars (note that errors in this plot are highly
        # correlated)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('normalized cdelt1')
        ax.set_ylabel('normalized crval1')
        ax.errorbar(cdelt1_search_norm, crval1_search_norm,
                    xerr=error_cdelt1_search_norm,
                    yerr=error_crval1_search_norm,
                    fmt='none')
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.plot(xp_limits, yp_limits, linestyle='-', color='magenta')
        ax.set_title("Potential solutions within the valid parameter space")
        # set window geometry
        set_window_geometry(geometry)
        print('Number of points in last plot:', len(cdelt1_search_norm))
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

    # ---
    # Segregate the different solutions (normalized to [0,1]) by
    # triplet. In this way the solutions are saved in different layers
    # (a layer for each triplet). The solutions will be stored as python
    # lists of numpy arrays.
    ntriplets_layered_list = []
    cdelt1_layered_list = []
    error_cdelt1_layered_list = []
    crval1_layered_list = []
    error_crval1_layered_list = []
    itriplet_layered_list = []
    clabel_layered_list = []
    for i in range(ntriplets_arc):
        ldum = (itriplet_search == i)
        ntriplets_layered_list.append(ldum.sum())
        #
        cdelt1_dum = cdelt1_search_norm[ldum]
        cdelt1_layered_list.append(cdelt1_dum)
        error_cdelt1_dum = error_cdelt1_search_norm[ldum]
        error_cdelt1_layered_list.append(error_cdelt1_dum)
        #
        crval1_dum = crval1_search_norm[ldum]
        crval1_layered_list.append(crval1_dum)
        error_crval1_dum = error_crval1_search_norm[ldum]
        error_crval1_layered_list.append(error_crval1_dum)
        #
        itriplet_dum = itriplet_search[ldum]
        itriplet_layered_list.append(itriplet_dum)
        #
        clabel_dum = [k for (k, v) in zip(clabel_search, ldum) if v]
        clabel_layered_list.append(clabel_dum)

    if abs(debugplot) >= 10:
        print('>>> Total number of potential solutions: ' +
              str(sum(ntriplets_layered_list)) + " (double check ==) " +
              str(len(itriplet_search)))
        print('>>> List with no. of solutions/triplet.:\n' +
              str(ntriplets_layered_list))
        pause_debugplot(debugplot)

    # ---
    # Computation of the cost function.
    #
    # For each solution, corresponding to a particular triplet, find
    # the nearest solution in each of the remaining ntriplets_arc-1
    # layers. Compute the distance (in normalized coordinates) to those
    # closest solutions, and obtain the sum of distances considering
    # only a fraction of them (after sorting them in ascending order).
    ntriplets_for_sum = max(
        1, int(round(frac_triplets_for_sum*float(ntriplets_arc)))
    )
    funcost_search = np.zeros(len(itriplet_search))
    for k in range(len(itriplet_search)):
        itriplet_local = itriplet_search[k]
        x0 = cdelt1_search_norm[k]
        y0 = crval1_search_norm[k]
        dist_to_layers = np.array([])
        for i in range(ntriplets_arc):
            if i != itriplet_local:
                if ntriplets_layered_list[i] > 0:
                    x1 = cdelt1_layered_list[i]
                    y1 = crval1_layered_list[i]
                    dist2 = (x0-x1)**2 + (y0-y1)**2
                    dist_to_layers = np.append(dist_to_layers, [min(dist2)])
                else:
                    dist_to_layers = np.append(dist_to_layers, [np.inf])
        dist_to_layers.sort()  # in-place sort
        funcost_search[k] = dist_to_layers[range(ntriplets_for_sum)].sum()

    # normalize the cost function
    funcost_min = min(funcost_search)
    if abs(debugplot) >= 10:
        print('funcost_min:', funcost_min)
    funcost_search /= funcost_min

    # segregate the cost function by arc triplet.
    funcost_layered_list = []
    for i in range(ntriplets_arc):
        ldum = (itriplet_search == i)
        funcost_dum = funcost_search[ldum]
        funcost_layered_list.append(funcost_dum)
    if abs(debugplot) >= 10:
        for i in range(ntriplets_arc):
            if ntriplets_layered_list[i] > 0:
                jdum = funcost_layered_list[i].argmin()
                print('>>>', i, funcost_layered_list[i][jdum],
                      clabel_layered_list[i][jdum],
                      cdelt1_layered_list[i][jdum],
                      crval1_layered_list[i][jdum])
            else:
                print('>>>', i, None, "(None, None, None)", None, None)
        pause_debugplot(debugplot)

    # intermediate plots
    if abs(debugplot) in [21, 22]:
        from numina.array.display.matplotlib_qt import plt

        # CDELT1 vs CRVAL1 diagram (normalized coordinates) with symbol
        # size proportional to the inverse of the cost function
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('normalized cdelt1')
        ax.set_ylabel('normalized crval1')
        ax.scatter(cdelt1_search_norm, crval1_search_norm,
                   s=2000/funcost_search, c=itriplet_search, alpha=0.2)
        xmin = -0.05
        xmax = 1.05
        ymin = -0.05
        ymax = 1.05
        xp_limits = np.array([0., 1., 0., 0.])
        yp_limits = np.array([1., 0., 0., 1.])
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.plot(xp_limits, yp_limits, linestyle='-', color='red')
        ax.set_title("Potential solutions within the valid parameter space\n" +
                     "[symbol size proportional to 1/(cost function)]")
        # set window geometry
        set_window_geometry(geometry)
        print('Number of points in last plot:', len(cdelt1_search_norm))
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

        # CDELT1 vs CRVAL1 diagram (normalized coordinates)
        # with symbol size proportional to the inverse of the cost
        # function and over-plotting triplet number
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('normalized cdelt1')
        ax.set_ylabel('normalized crval1')
        ax.scatter(cdelt1_search_norm, crval1_search_norm,
                   s=2000/funcost_search, c=itriplet_search, alpha=0.2)
        for i in range(len(itriplet_search)):
            ax.text(cdelt1_search_norm[i], crval1_search_norm[i],
                    str(int(itriplet_search[i])), fontsize=6)
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.plot(xp_limits, yp_limits, linestyle='-', color='red')
        ax.set_title("Potential solutions: arc line triplet number\n" +
                     "[symbol size proportional to 1/(cost function)]")
        # set window geometry
        set_window_geometry(geometry)
        print('Number of points in last plot:', len(cdelt1_search))
        pause_debugplot(debugplot, pltshow=True, tight_layout=True)

        # CDELT1 vs CRVAL1 diagram (normalized coordinates)
        # for i in range(ntriplets_arc):
        #     fig = plt.figure()
        #     ax = fig.add_subplot(111)
        #     ax.set_xlabel('normalized cdelt1')
        #     ax.set_ylabel('normalized crval1')
        #     xdum = cdelt1_layered_list[i]
        #     ydum = crval1_layered_list[i]
        #     sdum = 2000/funcost_layered_list[i]
        #     ax.scatter(xdum, ydum, s=sdum, alpha=0.8)
        #     ax.set_xlim(xmin, xmax)
        #     ax.set_ylim(ymin, ymax)
        #     ax.plot(xp_limits, yp_limits, linestyle='-', color='red')
        #     ax.set_title("Potential solutions: arc line triplet " + str(i) +
        #              " (from 0 to " + str(ntriplets_arc-1) + ")\n" +
        #              "[symbol size proportional to 1/(cost function)]")
        #     # set window geometry
        #     set_window_geometry(geometry)
        #     print('Number of points in last plot:', xdum.size)
        #     pause_debugplot(debugplot, pltshow=True, tight_layout=True)

    # ---
    # Line identification: several scenarios are considered.
    #
    # * Lines with three identifications:
    #   - Category A: the three identifications are identical. Keep the
    #     lowest value of the three cost functions.
    #   - Category B: two identifications are identical and one is
    #     different. Keep the line with two identifications and the
    #     lowest of the corresponding two cost functions.
    #   - Category C: the three identifications are different. Keep the
    #     one which is closest to a previously identified category B
    #     line. Use the corresponding cost function.
    #
    # * Lines with two identifications (second and penultimate lines).
    #   - Category D: the two identifications are identical. Keep the
    #     lowest cost function value.
    #
    # * Lines with only one identification (first and last lines).
    #   - Category E: the two lines next (or previous) to the considered
    #     line have been identified. Keep its cost function.
    #

    # We store the identifications of each line in a python list of
    # lists named diagonal_ids (which grows as the different triplets
    # are considered). A similar list of lists is also employed to
    # store the corresponding cost functions.
    # It is important to set the identification of the lines to None
    # when no valid master triplet has been associated to a given
    # arc line triplet.
    for i in range(ntriplets_arc):
        if ntriplets_layered_list[i] > 0:
            jdum = funcost_layered_list[i].argmin()
            k1, k2, k3 = clabel_layered_list[i][jdum]
            funcost_dum = funcost_layered_list[i][jdum]
        else:
            k1, k2, k3 = None, None, None
            funcost_dum = np.inf
        if i == 0:
            diagonal_ids = [[k1], [k2], [k3]]
            diagonal_funcost = [[funcost_dum], [funcost_dum], [funcost_dum]]
        else:
            diagonal_ids[i].append(k1)
            diagonal_ids[i+1].append(k2)
            diagonal_ids.append([k3])
            diagonal_funcost[i].append(funcost_dum)
            diagonal_funcost[i+1].append(funcost_dum)
            diagonal_funcost.append([funcost_dum])

    if abs(debugplot) >= 10:
        for i in range(nlines_arc):
            print(i, diagonal_ids[i], diagonal_funcost[i])
        pause_debugplot(debugplot)

    # The solutions are stored in a list of WavecalFeature instances.
    # Each WavecalFeature contains the following elements:
    # - line_ok: bool, indicates whether the line has been properly
    #   identified
    # - category: 'A','B','C','D','E',..., 'X'. Note that 'X' indicates
    #   that the line is still undefined.
    # - id: index of the line in the master table
    # - funcost: cost function associated the the line identification

    # initialize list_of_wvfeatures
    list_of_wvfeatures = []
    for i in range(nlines_arc):
        tmp_feature = WavecalFeature(line_ok=False,
                                     category='X',
                                     lineid=-1,
                                     funcost=np.inf,
                                     xpos=xpos_arc[i],
                                     ypos=0.0,
                                     peak=0.0,
                                     fwhm=0.0,
                                     reference=0.0)
        list_of_wvfeatures.append(tmp_feature)

    # set clipping window (in Angstrom)
    # note that potential lines with wavelengths outside the interval
    # [wvmin_clip, wvmax_clip] will be ignored
    if wvmin_useful is None:
        wvmin_clip = 0.0
    else:
        wvmin_clip = wvmin_useful
    if wvmax_useful is None:
        wvmax_clip = 1.0E10
    else:
        wvmax_clip = wvmax_useful

    # Category A lines
    for i in range(2, nlines_arc - 2):
        j1, j2, j3 = diagonal_ids[i]
        if j1 == j2 == j3 and j1 is not None:
            if wvmin_clip <= wv_master[j1] <= wvmax_clip:
                list_of_wvfeatures[i].line_ok = True
                list_of_wvfeatures[i].category = 'A'
                list_of_wvfeatures[i].lineid = j1
                list_of_wvfeatures[i].funcost = min(diagonal_funcost[i])
                list_of_wvfeatures[i].reference = wv_master[j1]

    if abs(debugplot) >= 10:
        print('\n* Including category A lines:')
        for i in range(nlines_arc):
            print(i, list_of_wvfeatures[i])
        pause_debugplot(debugplot)

    # Category B lines
    for i in range(2, nlines_arc - 2):
        if not list_of_wvfeatures[i].line_ok:
            j1, j2, j3 = diagonal_ids[i]
            f1, f2, f3 = diagonal_funcost[i]
            if j1 == j2 and j1 is not None:
                if max(f1, f2) < f3:
                    if wvmin_clip <= wv_master[j1] <= wvmax_clip:
                        list_of_wvfeatures[i].line_ok = True
                        list_of_wvfeatures[i].category = 'B'
                        list_of_wvfeatures[i].lineid = j1
                        list_of_wvfeatures[i].funcost = min(f1, f2)
                        list_of_wvfeatures[i].reference = wv_master[j1]
            elif j1 == j3 and j1 is not None:
                if max(f1, f3) < f2:
                    if wvmin_clip <= wv_master[j1] <= wvmax_clip:
                        list_of_wvfeatures[i].line_ok = True
                        list_of_wvfeatures[i].category = 'B'
                        list_of_wvfeatures[i].lineid = j1
                        list_of_wvfeatures[i].funcost = min(f1, f3)
                        list_of_wvfeatures[i].reference = wv_master[j1]
            elif j2 == j3 and j2 is not None:
                if max(f2, f3) < f1:
                    if wvmin_clip <= wv_master[j2] <= wvmax_clip:
                        list_of_wvfeatures[i].line_ok = True
                        list_of_wvfeatures[i].category = 'B'
                        list_of_wvfeatures[i].lineid = j2
                        list_of_wvfeatures[i].funcost = min(f2, f3)
                        list_of_wvfeatures[i].reference = wv_master[j2]

    if abs(debugplot) >= 10:
        print('\n* Including category B lines:')
        for i in range(nlines_arc):
            print(i, list_of_wvfeatures[i])
        pause_debugplot(debugplot)

    # Category C lines
    for i in range(2, nlines_arc - 2):
        if not list_of_wvfeatures[i].line_ok:
            j1, j2, j3 = diagonal_ids[i]
            f1, f2, f3 = diagonal_funcost[i]
            if list_of_wvfeatures[i-1].category == 'B':
                if min(f2, f3) > f1:
                    if wvmin_clip <= wv_master[j1] <= wvmax_clip:
                        list_of_wvfeatures[i].line_ok = True
                        list_of_wvfeatures[i].category = 'C'
                        list_of_wvfeatures[i].lineid = j1
                        list_of_wvfeatures[i].funcost = f1
                        list_of_wvfeatures[i].reference = wv_master[j1]
            elif list_of_wvfeatures[i+1].category == 'B':
                if min(f1, f2) > f3:
                    if wvmin_clip <= wv_master[j3] <= wvmax_clip:
                        list_of_wvfeatures[i].line_ok = True
                        list_of_wvfeatures[i].category = 'C'
                        list_of_wvfeatures[i].lineid = j3
                        list_of_wvfeatures[i].funcost = f3
                        list_of_wvfeatures[i].reference = wv_master[j3]

    if abs(debugplot) >= 10:
        print('\n* Including category C lines:')
        for i in range(nlines_arc):
            print(i, list_of_wvfeatures[i])
        pause_debugplot(debugplot)

    # Category D lines
    for i in [1, nlines_arc - 2]:
        j1, j2 = diagonal_ids[i]
        if j1 == j2 and j1 is not None:
            if wvmin_clip <= wv_master[j1] <= wvmax_clip:
                f1, f2 = diagonal_funcost[i]
                list_of_wvfeatures[i].line_ok = True
                list_of_wvfeatures[i].category = 'D'
                list_of_wvfeatures[i].lineid = j1
                list_of_wvfeatures[i].funcost = min(f1, f2)
                list_of_wvfeatures[i].reference = wv_master[j1]

    if abs(debugplot) >= 10:
        print('\n* Including category D lines:')
        for i in range(nlines_arc):
            print(i, list_of_wvfeatures[i])
        pause_debugplot(debugplot)

    # Category E lines
    i = 0
    if list_of_wvfeatures[i+1].line_ok and list_of_wvfeatures[i+2].line_ok:
        j1 = diagonal_ids[i][0]
        if j1 is not None:
            if wvmin_clip <= wv_master[j1] <= wvmax_clip:
                list_of_wvfeatures[i].line_ok = True
                list_of_wvfeatures[i].category = 'E'
                list_of_wvfeatures[i].lineid = diagonal_ids[i][0]
                list_of_wvfeatures[i].funcost = diagonal_funcost[i][0]
                list_of_wvfeatures[i].reference = wv_master[j1]
    i = nlines_arc-1
    if list_of_wvfeatures[i-2].line_ok and list_of_wvfeatures[i-1].line_ok:
        j1 = diagonal_ids[i][0]
        if j1 is not None:
            if wvmin_clip <= wv_master[j1] <= wvmax_clip:
                list_of_wvfeatures[i].line_ok = True
                list_of_wvfeatures[i].category = 'E'
                list_of_wvfeatures[i].lineid = diagonal_ids[i][0]
                list_of_wvfeatures[i].funcost = diagonal_funcost[i][0]
                list_of_wvfeatures[i].reference = wv_master[j1]

    if abs(debugplot) >= 10:
        print('\n* Including category E lines:')
        for i in range(nlines_arc):
            print(i, list_of_wvfeatures[i])
        pause_debugplot(debugplot)
        fit_list_of_wvfeatures(list_of_wvfeatures, naxis1_arc, crpix1,
                               poly_degree_wfit, weighted=False,
                               geometry=geometry, debugplot=debugplot)

    # ---
    # Check that the solutions do not contain duplicated values. If
    # they are present (probably due to the influence of an unknown
    # line that unfortunately falls too close to a real line in the
    # master table), we keep the solution with the lowest cost
    # function. The removed lines are labelled as category='R'. The
    # procedure is repeated several times in case a line appears more
    # than twice.
    lduplicated = True
    nduplicated = 0
    while lduplicated:
        lduplicated = False
        for i1 in range(nlines_arc):
            if list_of_wvfeatures[i1].line_ok:
                j1 = list_of_wvfeatures[i1].lineid
                for i2 in range(i1+1, nlines_arc):
                    if list_of_wvfeatures[i2].line_ok:
                        j2 = list_of_wvfeatures[i2].lineid
                        if j1 == j2:
                            lduplicated = True
                            nduplicated += 1
                            f1 = list_of_wvfeatures[i1].funcost
                            f2 = list_of_wvfeatures[i2].funcost
                            if f1 < f2:
                                list_of_wvfeatures[i2].line_ok = False
                                list_of_wvfeatures[i2].category = 'R'
                                # do not uncomment the next line:
                                # list_of_wvfeatures[i2].reference = None
                            else:
                                list_of_wvfeatures[i1].line_ok = False
                                list_of_wvfeatures[i1].category = 'R'
                                # do not uncomment the next line:
                                # list_of_wvfeatures[i1].reference = None

    if abs(debugplot) >= 10:
        if nduplicated > 0:
            print('\n* Removing category R lines:')
            for i in range(nlines_arc):
                print(i, list_of_wvfeatures[i])
            fit_list_of_wvfeatures(list_of_wvfeatures, naxis1_arc, crpix1,
                                   poly_degree_wfit, weighted=False,
                                   geometry=geometry, debugplot=debugplot)
        else:
            print('\n* No duplicated category R lines have been found')

    # ---
    # Filter out points with a large deviation from a robust linear
    # fit. The filtered lines are labelled as category='T'.
    if abs(debugplot) >= 10:
        print('\n>>> Theil-Sen filtering...')
    nfit, ifit, xfit, yfit, wfit = select_data_for_fit(list_of_wvfeatures)
    if nfit < 5:
        nremoved = 0
        if abs(debugplot) >= 10:
            print("nfit=", nfit)
            print("=> Skipping Theil-Sen filtering!")
    else:
        intercept, slope = fit_theil_sen(xfit, yfit)
        if abs(debugplot) >= 10:
            cdelt1_approx = slope
            crval1_approx = intercept + slope * crpix1
            print('>>> Theil-Sen CRVAL1: ', crval1_approx)
            print('>>> Theil-Sen CDELT1: ', cdelt1_approx)
        rfit = yfit - (intercept + slope*xfit)
        if abs(debugplot) >= 10:
            print('rfit:\n', rfit)
        sigma_rfit = robust_std(rfit)
        if abs(debugplot) >= 10:
            print('robust std:', sigma_rfit)
            print('normal std:', np.std(rfit))
        nremoved = 0
        for i in range(nfit):
            if abs(rfit[i]) > times_sigma_theil_sen * sigma_rfit:
                list_of_wvfeatures[ifit[i]].line_ok = False
                list_of_wvfeatures[ifit[i]].category = 'T'
                # do not uncomment the next line:
                # list_of_wvfeatures[ifit[i]].reference = None
                nremoved += 1

    if abs(debugplot) >= 10:
        if nremoved > 0:
            print('\n* Removing category T lines:')
            for i in range(nlines_arc):
                print(i, list_of_wvfeatures[i])
            fit_list_of_wvfeatures(list_of_wvfeatures, naxis1_arc, crpix1,
                                   poly_degree_wfit, weighted=False,
                                   geometry=geometry, debugplot=debugplot)
        else:
            print('\nNo category T lines have been found and removed')

    # ---
    # Filter out points that deviates from a polynomial fit. The
    # filtered lines are labelled as category='P'.
    if times_sigma_polfilt > 0:
        if abs(debugplot) >= 10:
            print('\n>>> Polynomial filtering...')
        nfit, ifit, xfit, yfit, wfit = select_data_for_fit(list_of_wvfeatures)
        if nfit <= poly_degree_wfit:
            raise ValueError(
                f"Insufficient number of points for fit, nfit={nfit}")
        # Note: do not use weighted fit because the weights can be very
        # different and the fit is, in practice, forced to pass through
        # some points while ignoring other points. Sometimes this leads to
        # the rejection of valid points (especially at the borders).
        poly = Polynomial.fit(x=xfit, y=yfit, deg=poly_degree_wfit)
        poly = Polynomial.cast(poly)
        rfit = yfit - poly(xfit)
        if abs(debugplot) >= 10:
            print('rfit:', rfit)
        sigma_rfit = robust_std(rfit)
        if abs(debugplot) >= 10:
            print('robust std:', sigma_rfit)
            print('normal std:', np.std(rfit))
        nremoved = 0
        for i in range(nfit):
            if abs(rfit[i]) > times_sigma_polfilt * sigma_rfit:
                list_of_wvfeatures[ifit[i]].line_ok = False
                list_of_wvfeatures[ifit[i]].category = 'P'
                # do not uncomment the next line:
                # list_of_wvfeatures[ifit[i]].reference = None
                nremoved += 1

        if abs(debugplot) >= 10:
            if nremoved > 0:
                print('\n* Removing category P lines:')
                for i in range(nlines_arc):
                    print(i, list_of_wvfeatures[i])
                fit_list_of_wvfeatures(list_of_wvfeatures, naxis1_arc, crpix1,
                                       poly_degree_wfit, weighted=False,
                                       geometry=geometry, debugplot=debugplot)
            else:
                print('\nNo category P lines have been found and removed')
    else:
        if abs(debugplot) >= 10:
            print('\n=> Skipping polynomial filtering!')

    # ---
    # Remove outliers using the Cook distance. The filtered lines are
    # labelled as category='K'.
    if times_sigma_cook > 0:
        if abs(debugplot) >= 10:
            print('\n>>> Removing outliers using Cook distance...')
        nfit, ifit, xfit, yfit, wfit = select_data_for_fit(list_of_wvfeatures)
        # There must be enough points to compute reasonable Cook distances
        if nfit <= poly_degree_wfit + 3:
            nremoved = 0
            if abs(debugplot) >= 10:
                print("nfit=", nfit)
                print("=> Skipping outliers detection using Cook distance!")
        else:
            poly, yres, reject = polfit_residuals_with_cook_rejection(
                x=xfit, y=yfit, deg=poly_degree_wfit,
                times_sigma_cook=times_sigma_cook,
                geometry=geometry,
                debugplot=debugplot)
            nremoved = 0
            for i in range(nfit):
                if abs(reject[i]):
                    list_of_wvfeatures[ifit[i]].line_ok = False
                    list_of_wvfeatures[ifit[i]].category = 'K'
                    # do not uncomment the next line:
                    # list_of_wvfeatures[ifit[i]].reference = None
                    nremoved += 1

        if abs(debugplot) >= 10:
            if nremoved > 0:
                print('\n* Removing category K lines:')
                for i in range(nlines_arc):
                    print(i, list_of_wvfeatures[i])
                fit_list_of_wvfeatures(list_of_wvfeatures, naxis1_arc, crpix1,
                                       poly_degree_wfit, weighted=False,
                                       geometry=geometry, debugplot=debugplot)
            else:
                print('\nNo category K lines have been found and removed')
    else:
        if abs(debugplot) >= 10:
            print('\n=> Skipping outlier detection using Cook distance!')

    # ---
    # If all the arc lines have been identified, compute the final
    # fit and exit
    line_ok = np.array([wvfeature.line_ok for wvfeature in list_of_wvfeatures])
    if np.all(line_ok):
        return list_of_wvfeatures

    # ---
    # Include unidentified lines by using the prediction of the
    # polynomial fit to the current set of identified lines. The
    # included lines are labelled as category='I'.
    loop_include_new_lines = True
    new_lines_included = False
    while loop_include_new_lines:
        if abs(debugplot) >= 10:
            print('\n>>> Polynomial prediction of unknown lines...')
        nfit, ifit, xfit, yfit, wfit = select_data_for_fit(list_of_wvfeatures)
        if nfit <= poly_degree_wfit:
            raise ValueError("Insufficient number of points for fit.")
        poly = Polynomial.fit(x=xfit, y=yfit, deg=poly_degree_wfit)
        poly = Polynomial.cast(poly)
        rfit = yfit - poly(xfit)
        if abs(debugplot) >= 10:
            print('rfit:\n', rfit)
        sigma_rfit = robust_std(rfit)
        if abs(debugplot) >= 10:
            print('robust std:', sigma_rfit)
            print('normal std:', np.std(rfit))

        intercept, slope = fit_theil_sen(xfit, yfit)
        if abs(debugplot) >= 10:
            print('crval1, cdelt1 (linear fit):', intercept, slope)

        list_id_already_found = []
        list_funcost_already_found = []
        for i in range(nlines_arc):
            if list_of_wvfeatures[i].line_ok:
                list_id_already_found.append(list_of_wvfeatures[i].lineid)
                list_funcost_already_found.append(
                    list_of_wvfeatures[i].funcost)

        nnewlines = 0
        for i in range(nlines_arc):
            if not list_of_wvfeatures[i].line_ok:
                zfit = poly(xpos_arc[i])  # predicted wavelength
                isort = np.searchsorted(wv_master, zfit)
                if isort == 0:
                    ifound = 0
                    dlambda = wv_master[ifound]-zfit
                elif isort == nlines_master:
                    ifound = isort - 1
                    dlambda = zfit - wv_master[ifound]
                else:
                    dlambda1 = zfit-wv_master[isort-1]
                    dlambda2 = wv_master[isort]-zfit
                    if dlambda1 < dlambda2:
                        ifound = isort - 1
                        dlambda = dlambda1
                    else:
                        ifound = isort
                        dlambda = dlambda2
                if abs(debugplot) >= 10:
                    print(i, ifound, wv_master[ifound], zfit, dlambda)
                if ifound not in list_id_already_found:  # unused line
                    condition1 = dlambda < times_sigma_inclusion * sigma_rfit
                    condition2 = dlambda/slope < error_xpos_arc
                    if condition1 or condition2:
                        list_id_already_found.append(ifound)
                        list_of_wvfeatures[i].line_ok = True
                        list_of_wvfeatures[i].category = 'I'
                        list_of_wvfeatures[i].lineid = ifound
                        # assign the worse cost function value
                        list_of_wvfeatures[i].funcost = max(
                            list_funcost_already_found
                        )
                        list_of_wvfeatures[i].reference = wv_master[ifound]
                        nnewlines += 1

        if abs(debugplot) >= 10:
            if nnewlines > 0:
                new_lines_included = True
                print('\n* Including category I lines:')
                for i in range(nlines_arc):
                    print(i, list_of_wvfeatures[i])
                fit_list_of_wvfeatures(list_of_wvfeatures, naxis1_arc, crpix1,
                                       poly_degree_wfit, weighted=False,
                                       geometry=geometry, debugplot=debugplot)
            else:
                if new_lines_included:
                    print("\nNo additional category I lines have been found " +
                          "and added")
                else:
                    print('\nNo category I lines have been found and added')

        if nnewlines == 0:
            loop_include_new_lines = False

    return list_of_wvfeatures


def match_wv_arrays(wv_master, wv_expected_all_peaks, delta_wv_max):
    """Match two lists with wavelengths.

    Assign individual wavelengths from wv_master to each expected
    wavelength when the latter is within the maximum allowed range.

    Parameters
    ----------
    wv_master : numpy array
        Array containing the master wavelengths.
    wv_expected_all_peaks : numpy array
        Array containing the expected wavelengths (computed, for
        example, from an approximate polynomial calibration applied to
        the location of the line peaks).
    delta_wv_max : float
        Maximum distance to accept that the master wavelength
        corresponds to the expected wavelength.

    Returns
    -------
    wv_verified_all_peaks : numpy array
        Verified wavelengths from master list.

    """

    # initialize the output array to zero
    wv_verified_all_peaks = np.zeros_like(wv_expected_all_peaks)

    # initialize to True array to indicate that no peak has already
    # been verified (this flag avoids duplication)
    wv_unused = np.ones_like(wv_expected_all_peaks, dtype=bool)

    # initialize to np.inf array to store minimum distance to already
    # identified line
    minimum_delta_wv = np.ones_like(wv_expected_all_peaks, dtype=float)
    minimum_delta_wv *= np.inf

    # since it is likely that len(wv_master) < len(wv_expected_all_peaks),
    # it is more convenient to execute the search in the following order
    for i in range(len(wv_master)):
        j = np.searchsorted(wv_expected_all_peaks, wv_master[i])
        if j == 0:
            delta_wv = abs(wv_master[i] - wv_expected_all_peaks[j])
            if delta_wv < delta_wv_max:
                if wv_unused[j]:
                    wv_verified_all_peaks[j] = wv_master[i]
                    wv_unused[j] = False
                    minimum_delta_wv[j] = delta_wv
                else:
                    if delta_wv < minimum_delta_wv[j]:
                        wv_verified_all_peaks[j] = wv_master[i]
                        minimum_delta_wv[j] = delta_wv
        elif j == len(wv_expected_all_peaks):
            delta_wv = abs(wv_master[i] - wv_expected_all_peaks[j-1])
            if delta_wv < delta_wv_max:
                if wv_unused[j-1]:
                    wv_verified_all_peaks[j-1] = wv_master[i]
                    wv_unused[j-1] = False
                else:
                    if delta_wv < minimum_delta_wv[j-1]:
                        wv_verified_all_peaks[j-1] = wv_master[i]
        else:
            delta_wv1 = abs(wv_master[i] - wv_expected_all_peaks[j-1])
            delta_wv2 = abs(wv_master[i] - wv_expected_all_peaks[j])
            if delta_wv1 < delta_wv2:
                if delta_wv1 < delta_wv_max:
                    if wv_unused[j-1]:
                        wv_verified_all_peaks[j-1] = wv_master[i]
                        wv_unused[j-1] = False
                    else:
                        if delta_wv1 < minimum_delta_wv[j-1]:
                            wv_verified_all_peaks[j-1] = wv_master[i]
            else:
                if delta_wv2 < delta_wv_max:
                    if wv_unused[j]:
                        wv_verified_all_peaks[j] = wv_master[i]
                        wv_unused[j] = False
                    else:
                        if delta_wv2 < minimum_delta_wv[j]:
                            wv_verified_all_peaks[j] = wv_master[i]

    return wv_verified_all_peaks


def refine_arccalibration(sp, poly_initial, wv_master, poldeg,
                          nrepeat=3,
                          ntimes_match_wv=2,
                          nwinwidth_initial=7,
                          nwinwidth_refined=5,
                          times_sigma_reject=5,
                          interactive=False,
                          threshold=0,
                          plottitle=None,
                          decimal_places=4,
                          ylogscale=False,
                          geometry=None,
                          pdf=None,
                          debugplot=0):
    """Refine wavelength calibration using an initial polynomial.

    Parameters
    ----------
    sp : numpy array
        1D array of length NAXIS1 containing the input spectrum.
    poly_initial : Polynomial instance
        Initial wavelength calibration polynomial, providing the
        wavelength as a function of pixel number (running from 1 to
        NAXIS1).
    wv_master : numpy array
        Array containing the master list of arc line wavelengths.
    poldeg : int
        Polynomial degree of refined wavelength calibration. Note
        that this degree can be different from the polynomial degree
        of poly_initial.
    nrepeat : int
        Number of times lines are iteratively included in the initial
        fit.
    ntimes_match_wv : int
        Number of pixels around each line peak where the expected
        wavelength must match the tabulated wavelength in the master
        list.
    nwinwidth_initial : int
        Initial window width to search for line peaks in spectrum.
    nwinwidth_refined : int
        Window width to refine line peak location.
    times_sigma_reject : float
        Times sigma to reject points in the fit.
    interactive : bool
        If True, the function allows the user to modify the fit
        interactively.
    threshold : float
        Minimum signal in the peaks.
    plottitle : string or None
        Plot title.
    decimal_places : int
        Number of decimal places to be employed when displaying relevant
        fitted parameters.
    ylogscale : bool
        If True, the spectrum is displayed in logarithmic units. Note
        that this is only employed for display purposes. The line peaks
        are found in the original spectrum.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    pdf : PdfFile object or None
        If not None, output is sent to PDF file.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    poly_refined : Polynomial instance
        Refined wavelength calibration polynomial.
    yres_summary : dictionary
        Statistical summary of the residuals.

    """

    # check that nrepeat is larger than zero
    if nrepeat <= 0:
        raise ValueError("Unexpected nrepeat=", str(nrepeat))

    # check that the requested polynomial degree is equal or larger than
    # the degree of the initial polynomial
    if poldeg < len(poly_initial.coef) - 1:
        raise ValueError("Polynomial degree of refined polynomial must be "
                         "equal or larger than that of the initial polynomial")

    # check that interactive use takes place when plotting
    if interactive:
        if abs(debugplot) % 10 != 0:
            local_debugplot = debugplot
        else:
            local_debugplot = -12
    else:
        local_debugplot = debugplot

    local_ylogscale = ylogscale

    # latest limits
    global xmin_previous
    global xmax_previous
    global ymin_previous
    global ymax_previous
    xmin_previous = None
    xmax_previous = None
    ymin_previous = None
    ymax_previous = None

    # spectrum length
    naxis1 = sp.shape[0]

    # define default values in case no useful lines are identified
    fxpeaks = np.array([])
    poly_refined = np.polynomial.Polynomial([0.0])
    poldeg_effective = len(poly_refined.coef) - 1
    yres_summary = summary(np.array([]))

    # compute linear values from initial polynomial
    crmin1_linear = poly_initial(1)
    crmax1_linear = poly_initial(naxis1)
    cdelt1_linear = (crmax1_linear - crmin1_linear) / (naxis1 - 1)

    # find initial line peaks
    ixpeaks = find_peaks_spectrum(sp,
                                  nwinwidth=nwinwidth_initial,
                                  threshold=threshold)
    npeaks = len(ixpeaks)

    if npeaks > 0:

        # refine line peak locations
        fxpeaks, sxpeaks = refine_peaks_spectrum(
            sp, ixpeaks,
            nwinwidth=nwinwidth_refined,
            method="gaussian"
        )

        # expected wavelength of all identified peaks
        wv_expected_all_peaks = poly_initial(fxpeaks + 1.0)

        # assign individual arc lines from master list to spectrum
        # line peaks when the expected wavelength is within the maximum
        # allowed range (+/- ntimes_match_wv * CDELT1 around the peak)
        delta_wv_max = ntimes_match_wv * cdelt1_linear
        wv_verified_all_peaks = match_wv_arrays(
            wv_master,
            wv_expected_all_peaks,
            delta_wv_max=delta_wv_max
        )

    loop = True

    nrepeat_eff = nrepeat
    while loop:

        nlines_ok = 0
        xdum = np.array([])
        ydum = np.array([])
        reject = np.array([])

        if npeaks > 0:

            for irepeat in range(nrepeat_eff):
                # fit with sigma rejection
                lines_ok = np.where(wv_verified_all_peaks > 0)
                nlines_ok = len(lines_ok[0])

                # there are matched lines
                if nlines_ok > 0:
                    # select points to be fitted
                    xdum = (fxpeaks + 1.0)[lines_ok]
                    ydum = wv_verified_all_peaks[lines_ok]

                    # determine effective polynomial degree
                    if nlines_ok > poldeg:
                        poldeg_effective = poldeg
                    else:
                        poldeg_effective = nlines_ok - 1

                    # fit polynomial
                    poly_refined, yres, reject = \
                        polfit_residuals_with_sigma_rejection(
                            x=xdum,
                            y=ydum,
                            deg=poldeg_effective,
                            times_sigma_reject=times_sigma_reject,
                            debugplot=0
                        )

                    # effective number of points
                    yres_summary = summary(yres[np.logical_not(reject)])

                else:
                    poly_refined = np.polynomial.Polynomial([0.0])
                    yres_summary = summary(np.array([]))

                if irepeat < nrepeat_eff - 1:
                    delta_wv_max = ntimes_match_wv * cdelt1_linear
                    wv_verified_all_peaks = match_wv_arrays(
                        wv_master,
                        poly_refined(fxpeaks + 1.0),
                        delta_wv_max=delta_wv_max
                    )

        # update poldeg_effective
        poldeg_effective = len(poly_refined.coef) - 1

        # update linear values
        crpix1_linear = 1.0
        crval1_linear = poly_refined(crpix1_linear)
        crmin1_linear = poly_refined(1)
        crmax1_linear = poly_refined(naxis1)
        cdelt1_linear = (crmax1_linear - crmin1_linear) / (naxis1 - 1)

        if abs(local_debugplot) >= 10:
            print(79 * '=')
            print(">>> poldeg (requested, effective)..:",
                  poldeg, poldeg_effective)
            print(">>> Fitted coefficients............:\n", poly_refined.coef)
            print(">>> NAXIS1.........................:", naxis1)
            print(">>> CRVAL1 linear scale............:", crval1_linear)
            print(">>> CDELT1 linear scale............:", cdelt1_linear)
            print(79 * '.')
            print(">>> Number of peaks................:", npeaks)
            print(">>> nlines identified (total, used): ",
                  nlines_ok, yres_summary['npoints'])
            print(">>> robust_std.....................:",
                  yres_summary['robust_std'])
            print(79 * '-')

        if (abs(local_debugplot) % 10 != 0) or (pdf is not None):
            from numina.array.display.matplotlib_qt import plt

            def handle_close(evt):
                global xmin_previous
                global xmax_previous
                global ymin_previous
                global ymax_previous
                xmin_previous, xmax_previous = ax2.get_xlim()
                ymin_previous, ymax_previous = ax2.get_ylim()

            if pdf is not None:
                fig = plt.figure(figsize=(11.69, 8.27), dpi=100)
            else:
                fig = plt.figure()
                fig.canvas.mpl_connect('close_event', handle_close)
            set_window_geometry(geometry)

            grid = plt.GridSpec(2, 1)
            grid.update(left=0.10, right=0.98,
                        bottom=0.10, top=0.90, hspace=0.01)

            # differences between linear fit and fitted polynomial
            # polynomial fit
            xpol = np.arange(1, naxis1 + 1)
            ylinear = crval1_linear + (xpol - crpix1_linear) * cdelt1_linear
            ypol = poly_refined(xpol) - ylinear
            # identified lines
            yp = ydum - (crval1_linear + (xdum - crpix1_linear) *
                         cdelt1_linear)

            # upper plot
            ax1 = fig.add_subplot(grid[0, 0])
            if nlines_ok > 0:
                ymin = min(ypol)
                ymax = max(ypol)
                dy = ymax - ymin
                if dy > 0:
                    ymin -= dy/50
                    ymax += dy/50
                else:
                    ymin -= 0.5
                    ymax += 0.5
            else:
                ymin = -1.0
                ymax = 1.0
            ax1.set_xlim(1 - 0.02 * naxis1, naxis1 * 1.02)
            ax1.set_ylim(ymin, ymax)
            if nlines_ok > 0:
                ax1.plot(xdum, yp, 'mo', label='identified')
                if sum(reject) > 0:
                    ax1.plot(xdum[reject], yp[reject], 'o',
                             color='tab:gray', label='ignored')
                ax1.plot(xpol, ypol, 'c-', label='fit')
            ax1.set_ylabel('polynomial - linear fit ' + r'($\AA$)')
            ax1.text(0.01, 0.99, 'CRVAL1 (' + r'$\AA$' + '):' +
                     str(round(crval1_linear, decimal_places)),
                     horizontalalignment='left',
                     verticalalignment='top',
                     transform=ax1.transAxes)
            ax1.text(0.01, 0.91, 'CDELT1 (' + r'$\AA$' + '/pixel):' +
                     str(round(cdelt1_linear, decimal_places)),
                     horizontalalignment='left',
                     verticalalignment='top',
                     transform=ax1.transAxes)
            ax1.text(0.99, 0.99, 'robust std (' + r'$\AA$' + '):' +
                     str(round(yres_summary['robust_std'], decimal_places)),
                     horizontalalignment='right',
                     verticalalignment='top',
                     transform=ax1.transAxes)
            ax1.text(0.5, 0.55, 'No. points (total / used): ' +
                     str(nlines_ok) + ' / ' +
                     str(yres_summary['npoints']),
                     horizontalalignment='center',
                     verticalalignment='top',
                     transform=ax1.transAxes)
            ax1.text(0.5, 0.4, 'Polynomial degree: ' + str(poldeg_effective),
                     horizontalalignment='center',
                     verticalalignment='bottom',
                     transform=ax1.transAxes)
            if plottitle is None:
                ax1.set_title('Refined wavelength calibration')
            else:
                ax1.set_title(plottitle)
            ax1.legend(numpoints=1, ncol=3, fancybox=True,
                       loc='lower center')
            #           bbox_to_anchor=(0.5, 1.00))

            # lower plot
            if local_ylogscale:
                spectrum = sp - sp.min() + 1.0
                spectrum = np.log10(spectrum)
            else:
                spectrum = sp.copy()
            ax2 = fig.add_subplot(grid[1, 0], sharex=ax1)
            ax2.set_xlim(1 - 0.02 * naxis1, naxis1 * 1.02)
            if local_ylogscale:
                ymin = spectrum[ixpeaks].min()
            else:
                ymin = spectrum.min()
            ymax = spectrum.max()
            dy = ymax - ymin
            ymin -= dy / 40.
            ymax += dy / 40.
            ax2.set_ylim(ymin, ymax)
            if xmin_previous is not None:
                ax2.set_xlim(xmin_previous, xmax_previous)
                ax2.set_ylim(ymin_previous, ymax_previous)
            ax2.plot(xpol, spectrum, '-')
            ax2.set_xlabel('pixel position (from 1 to NAXIS1)')
            if local_ylogscale:
                ax2.set_ylabel('~ log10(number of counts)')
            else:
                ax2.set_ylabel('number of counts')
            # mark peak location
            # ax2.plot(ixpeaks + 1, spectrum[ixpeaks], 'co',
            #          label="initial location")
            # ax2.plot(fxpeaks + 1, spectrum[ixpeaks], 'go',
            #          label="refined location")
            ax2.plot((fxpeaks + 1)[lines_ok], spectrum[ixpeaks][lines_ok],
                     'mo', label="identified lines")
            for i in range(len(ixpeaks)):
                if wv_verified_all_peaks[i] > 0:
                    ax2.text(fxpeaks[i] + 1.0, spectrum[ixpeaks[i]],
                             str(wv_verified_all_peaks[i]) +
                             '(' + str(i + 1) + ')',
                             fontsize=8,
                             horizontalalignment='center')
                else:
                    ax2.text(fxpeaks[i] + 1.0, spectrum[ixpeaks[i]],
                             '(' + str(i + 1) + ')',
                             fontsize=8,
                             horizontalalignment='center')
            # display expected location of lines in master file
            for i in range(len(wv_master)):
                tempol = poly_refined.copy()
                tempol.coef[0] -= wv_master[i]
                # compute roots
                tmproots = tempol.roots()
                # select real solutions
                tmproots = tmproots.real[abs(tmproots.imag) < 1e-5]
                # choose values within valid channel range
                tmproots = tmproots[(tmproots >= 1) * (tmproots <= naxis1)]
                if len(tmproots) > 0:
                    ax2.plot([tmproots[0], tmproots[0]], [ymin, ymax],
                             color='grey', linestyle='dotted')
            # legend
            ax2.legend(numpoints=1)

            if pdf is not None:
                pdf.savefig()
            else:
                if local_debugplot in [-22, -12, 12, 22]:
                    pause_debugplot(
                        debugplot=local_debugplot,
                        optional_prompt='Zoom/Unzoom or ' +
                                        'press RETURN to continue...',
                        tight_layout=False,
                        pltshow=True
                    )
                else:
                    pause_debugplot(debugplot=local_debugplot,
                                    tight_layout=False, pltshow=True)

            # request next action in interactive session
            if interactive:
                nrepeat_eff = 1
                print('Recalibration menu')
                print('------------------')
                print('[i] (i)nsert new peak and restart')
                print('[d] (d)elete all the identified lines')
                print('[r] (r)estart from begining')
                print('[a] (a)utomatic line inclusion')
                print('[l] toggle (l)ogarithmic scale on/off')
                print('[e] (e)valuate current polynomial at a given pixel')
                print('[w] replot (w)hole spectrum')
                print('[x] e(x)it without additional changes')
                print('[#] from 1 to ' + str(len(ixpeaks)) +
                      ' --> modify line #')
                ioption = readi('Option', default='x',
                                minval=1, maxval=len(ixpeaks),
                                allowed_single_chars='adeilrwx')
                if ioption == 'd':
                    wv_verified_all_peaks = np.zeros(npeaks)
                elif ioption == 'r':
                    delta_wv_max = ntimes_match_wv * cdelt1_linear
                    wv_expected_all_peaks = poly_initial(fxpeaks + 1.0)
                    wv_verified_all_peaks = match_wv_arrays(
                        wv_master,
                        wv_expected_all_peaks,
                        delta_wv_max=delta_wv_max
                    )
                elif ioption == 'a':
                    delta_wv_max = ntimes_match_wv * cdelt1_linear
                    wv_verified_all_peaks = match_wv_arrays(
                        wv_master,
                        poly_refined(fxpeaks + 1.0),
                        delta_wv_max=delta_wv_max
                    )
                elif ioption == 'l':
                    xmin_previous = None
                    xmax_previous = None
                    ymin_previous = None
                    ymax_previous = None
                    if local_ylogscale:
                        local_ylogscale = False
                    else:
                        local_ylogscale = True
                elif ioption == 'e':
                    pixel = 1
                    while pixel != 0:
                        pixel = readf("Pixel coordinate (0=exit)",
                                      default=0)
                        print("--> Wavelength:", poly_refined(pixel))
                elif ioption == 'i':
                    ipixel = 1
                    # include new peaks
                    while ipixel != 0:
                        ipixel = readi("Closest pixel coordinate (integer) "
                                       "to insert peak (0=exit)",
                                       default=0, minval=0, maxval=naxis1)
                        if ipixel > 0:
                            ixpeaks = np.concatenate((ixpeaks,
                                                      np.array([ipixel-1])))
                    # sort updated array
                    ixpeaks.sort()
                    npeaks = len(ixpeaks)
                    # refine line peak locations
                    fxpeaks, sxpeaks = refine_peaks_spectrum(
                        sp, ixpeaks,
                        nwinwidth=nwinwidth_refined,
                        method="gaussian"
                    )
                    # expected wavelength of all identified peaks
                    delta_wv_max = ntimes_match_wv * cdelt1_linear
                    wv_verified_all_peaks = match_wv_arrays(
                        wv_master,
                        poly_refined(fxpeaks + 1.0),
                        delta_wv_max=delta_wv_max
                    )
                elif ioption == 'w':
                    xmin_previous = None
                    xmax_previous = None
                    ymin_previous = None
                    ymax_previous = None
                elif ioption == 'x':
                    loop = False
                else:
                    print(wv_master)
                    expected_value = \
                        poly_refined(fxpeaks[ioption - 1] + 1.0)
                    print('>>> Current expected wavelength for line #' +
                          str(ioption) + ": ", expected_value)
                    delta_wv_max = ntimes_match_wv * cdelt1_linear
                    close_value = match_wv_arrays(
                        wv_master,
                        np.array([expected_value]),
                        delta_wv_max=delta_wv_max)
                    newvalue = readf('New value for line #' + str(ioption) +
                                     ' (0 to delete line)',
                                     default=close_value[0])
                    wv_verified_all_peaks[ioption - 1] = newvalue

            else:
                loop = False

        else:
            loop = False

    # if effective degree of poly_refined < poldeg, add zeros
    if poldeg_effective < poldeg:
        numzeros = poldeg - poldeg_effective
        final_coefficients = np.concatenate((poly_refined.coef,
                                             np.zeros(numzeros)))
        poly_refined = np.polynomial.Polynomial(final_coefficients)

    if abs(local_debugplot) >= 10:
        print(">>> Initial coefficients:\n", poly_initial.coef)
        print(">>> Refined coefficients:")
        for cdum in poly_refined.coef:
            print(cdum)
        print(">>> Final CRVAL1 linear scale............:", crval1_linear)
        print(">>> Final CDELT1 linear scale............:", cdelt1_linear)

    return poly_refined, yres_summary



import pytest


def test_wheel(pattri_ins):
    dev = pattri_ins.get_device("wheel")
    assert dev.get_property("position") == 0
    dev.turn()
    assert dev.get_property("position") == 1
    assert dev.get_property("label") == "B"
    pattri_ins.get_device("wheel").position = 3
    assert dev.get_property("position") == 3
    assert dev.get_property("label") == "R"
    assert dev._capacity == 5


def test_wheel_proxy1(pattri_ins, pattri_header):
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_ins.get_property("wheel.position") == 3
    assert pattri_ins.get_property("pos") == 3
    pattri_ins.pos = 4
    assert pattri_ins.get_property("wheel.position") == 4
    pattri_ins.get_device("wheel").position = 0
    assert pattri_ins.pos == 0
    pattri_ins.pos = 2
    assert pattri_ins.pos == 2
    assert pattri_ins.get_property("wheel.position") == 2


def test_wheel_proxy2(pattri_ins):
    assert pattri_ins.get_property("wheel.position") == 0
    assert pattri_ins.get_property("pos") == 0
    pattri_ins.pos = 2
    assert pattri_ins.get_property("wheel.position") == 2
    assert pattri_ins.get_device("wheel").position == 2


def test_box_proxy2(pattri_ins):
    assert pattri_ins.get_property("PSU.mode") == "Mode_A"
    assert pattri_ins.get_property("mode") == "Mode_A"
    with pytest.raises(ValueError):
        pattri_ins.mode = "R"

    with pytest.raises(ValueError):
        pattri_ins.get_device("PSU").mode = "M"



#
# Copyright 2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Extract 2D slice from 3D FITS file."""

from astropy.io import fits
import argparse
from datetime import datetime
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import sys


def extract_slice(input, axis, i1, i2, method, wavecal, transpose, vmin, vmax, noplot, output):
    """Extract 2D slice.

    Parameters
    ----------
    input : str
        Input FITS file name.
    axis : int
        Axis to be collapsed in output.
    i1 : int
        First pixel of the projected axis (FITS criterum).
    i2 : int
        Last pixel of the projected axis (FITS criterium).
    method : str
        Collapse method. Choices are "sum", "mean", "median"
    wavecal : str
        Wavelength calibration type. Choices are:
        - "none": ignore wavelength calibration parameters
        - "fridasimulator": the CDELT3 value is 1.0 and the
          wavelength increment must be read from PC3_3
        - "1", "2" or "3": read the traditional CTYPE?, CRPIX?, CRVAL?,
          and CDELT?, with ? = 1, 2 or 3
    transpose : bool
        If True, transpose data array in output.
    vmin : float or None
        Minimum data range that the colormap covers.
    vmax : float or None
        Maximum data range that the colormap covers.
    noplot : bool
        If True, skip plotting of the result.
    output : str
        Output FITS file name.

    """

    # protections
    if not (1 <= axis <= 3):
        raise ValueError(f'{axis=} out of valid range (1, 2 or 3)')

    # read first FITS file
    with fits.open(input) as hdulist:
        header = hdulist[0].header
        data = hdulist[0].data.astype(float)
    naxis = header['naxis']
    if naxis != 3:
        raise ValueError(f'Unexpected input {naxis=} (it must be 3)')
    naxis3 = header[f'naxis{axis}']

    # set last pixel to maximum value when i2==0
    if i2 == 0:
        i2 = naxis3

    if not (1 <= i1 <= naxis3):
        raise ValueError(f'Unexpected {i1=}. Must be a number in [1,...,{naxis3}]')
    if not (1 <= i2 <= naxis3):
        raise ValueError(f'Unexpected {i2=}. Must be a number in [1,...,{naxis3}]')
    if i2 < i1:
        raise ValueError(f'{i1=} must be <= {i2=}')

    if axis == 1:
        if np.any(np.isnan(data[:, :, (i1 - 1):i2])):
            print(f'Warning: NaN values found in input data for axis {axis} '
                  f'between pixels {i1} and {i2}. They will be ignored.')
        if method == "sum":
            slice2d = np.nansum(data[:, :, (i1 - 1):i2], axis=2, keepdims=False)
        elif method == "mean":
            slice2d = np.nanmean(data[:, :, (i1 - 1):i2], axis=2, keepdims=False)
        elif method == "median":
            slice2d = np.nanmedian(data[:, :, (i1 - 1):i2], axis=2, keepdims=False)
        else:
            raise ValueError(f'Unexpected {method=}')
    elif axis == 2:
        if np.any(np.isnan(data[:, (i1 - 1):i2, :])):
            print(f'Warning: NaN values found in input data for axis {axis} '
                  f'between pixels {i1} and {i2}. They will be ignored.')
        if method == "sum":
            slice2d = np.nansum(data[:, (i1 - 1):i2, :], axis=1, keepdims=False)
        elif method == "mean":
            slice2d = np.nanmean(data[:, (i1 - 1):i2, :], axis=1, keepdims=False)
        elif method == "median":
            slice2d = np.nanmedian(data[:, (i1 - 1):i2, :], axis=1, keepdims=False)
        else:
            raise ValueError(f'Unexpected {method=}')
    else:
        if np.any(np.isnan(data[(i1 - 1):i2, :, :])):
            print(f'Warning: NaN values found in input data for axis {axis} '
                  f'between pixels {i1} and {i2}. They will be ignored.')
        if method == "sum":
            slice2d = np.nansum(data[(i1 - 1):i2, :, :], axis=0, keepdims=False)
        elif method == "mean":
            slice2d = np.nanmean(data[(i1 - 1):i2, :, :], axis=0, keepdims=False)
        elif method == "median":
            slice2d = np.nanmedian(data[(i1 - 1):i2, :, :], axis=0, keepdims=False)
        else:
            raise ValueError(f'Unexpected {method=}')

    if transpose:
        slice2d = np.transpose(slice2d)

    # plot result
    if not noplot:
        naxis2, naxis1 = slice2d.shape
        if wavecal != 'none':
            aspect = 'auto'
        else:
            aspect = None
        xmin, xmax = 0.5, naxis1 + 0.5
        ymin, ymax = 0.5, naxis2 + 0.5
        extent = [xmin, xmax, ymin, ymax]
        fig, ax = plt.subplots()
        img = ax.imshow(slice2d, origin='lower', extent=extent, vmin=vmin, vmax=vmax,
                        aspect=aspect, interpolation='None')
        ax.set_xlabel('pixel (output NAXIS1)')
        ax.set_ylabel('pixel (output NAXIS2)')
        title = f'{input}\n(collapsed input NAXIS{axis} [{i1}, {i2}])'
        ax.set_title(title)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(img, cax=cax, label=f'Number of counts ({method})')
        plt.tight_layout()
        plt.show()

    # save result
    if output is not None:
        hdu = fits.PrimaryHDU(slice2d.astype(np.float32))
        header_output = hdu.header

        if wavecal != "none":
            if wavecal in "123":
                axiswave = int(wavecal)
                header_output['ctype1'] = header[f'ctype{axiswave}']
                header_output['crpix1'] = header[f'crpix{axiswave}']
                header_output['crval1'] = header[f'crval{axiswave}']
                header_output['cdelt1'] = header[f'cdelt{axiswave}']
            elif wavecal == "fridasimulator":
                axiswave = 3
                header_output['ctype1'] = header[f'ctype{axiswave}']
                header_output['crpix1'] = header[f'crpix{axiswave}']
                header_output['crval1'] = header[f'crval{axiswave}']
                header_output['cdelt1'] = header[f'pc{axiswave}_{axiswave}']
            else:
                raise ValueError(f"Unexpected {wavecal=}")

        header_output['history'] = f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
        header_output['comment'] = 'Image created with extract_2d_slice_from_3d_cube.py'
        header_output['comment'] = 'with the following arguments:'
        header_output['comment'] = f'input: {input}'
        header_output['comment'] = f'axis: {axis}'
        header_output['comment'] = f'i1: {i1}'
        header_output['comment'] = f'i2: {i2}'
        header_output['comment'] = f'--method: {method}'
        header_output['comment'] = f'--wavecal: {wavecal}'
        header_output['comment'] = f'--transpose: {transpose}'
        header_output['comment'] = f'output: {output}'

        hdu.writeto(output, overwrite="yes")


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="Input FITS file")
    parser.add_argument("--axis", help="Axis to be collapsed in output", type=int, default=3)
    parser.add_argument("--i1", help="First pixel of the projected axis", type=int, default=1)
    parser.add_argument("--i2", help="Last pixel of the projected axis (0=NAXIS value)", type=int, default=0)
    parser.add_argument("--method", help="Collapse method (default=sum)", type=str, default="sum",
                        choices=["sum", "mean", "median"])
    parser.add_argument("--wavecal", help="Wavelength calibration type", type=str, default="none",
                        choices=["none", "fridasimulator", "1", "2", "3"])
    parser.add_argument("--transpose", help="Transpose data array in output", action="store_true")
    parser.add_argument("--noplot", help="Do not plot result", action="store_true")
    parser.add_argument("--vmin", help="vmin value for imshow", type=float)
    parser.add_argument("--vmax", help="vmax value for imshow", type=float)
    parser.add_argument("--output", help="Output FITS file")
    parser.add_argument("--echo", help="Display full command line", action="store_true")
    parser.add_argument("--debug", help="Debug", action="store_true")

    args = parser.parse_args(args=args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.debug:
        for arg, value in vars(args).items():
            print(f'{arg}: {value}')

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    extract_slice(
        input=args.input,
        axis=args.axis,
        i1=args.i1,
        i2=args.i2,
        method=args.method,
        wavecal=args.wavecal,
        transpose=args.transpose,
        vmin=args.vmin,
        vmax=args.vmax,
        noplot=args.noplot,
        output=args.output
    )


if __name__ == "__main__":

    main()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
from astropy.coordinates import SkyCoord, Angle
from astropy.units import Unit, Quantity
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
import numpy as np


def plot_north_east_arrows(
        ax,
        wcs2d,
        arrow_length=None,
        scale_length=None,
        scale_location=1,
        parangle=None,
        fits_criterion=False,
        color='grey',
        color_scale=None,
        color_parangle=None,
        fontsize_scale_relative_factor=0.8,
        verbose=False
):
    """Display North & East arrow using WCS information.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Instance of `matplotlib.axes.Axes`.
    wcs2d : astropy.wcs.WCS
        Instance of a 2D `astropy.wcs.WCS`.
    arrow_length : astropy.coordinates.Angle, optional
        Arrow length with Angle units.
    scale_length : Astropy.coordinates.Angle
        Size of the scale indication, with Angle units.
        If None, 'arrow_length' is assumed.
    scale_location : int
        Location of the scale (with the same length as the NE arrows).
        0: do not plot ruler
        1: upper right
        2: upper left
        3: lower left
        4: lower right
    parangle : Astropy.coordinates.Angle
        Parallactic angle with Angle units. Displayed only if this
        parameter is given.
    fits_criterion : bool
        If True, assume FITS criteria to compute the central pixel.
    color : str
        Color for plotting the NE arrows.
    color_scale : str or None
        Color for plotting the scale segment.
        If None, 'color' is employed.
    color_parangle : str or None
        Color for parangle line. If none, 'color' is employed.
    fontsize_scale_relative_factor : float
        Scaling factor to determine the font size employed in the
        scale. A value of 1.0 indicates that the font size is the
        same as the one employed in the plot labels.
    verbose : bool
        If True, display additional information.
    """

    # protections
    if arrow_length is not None:
        if not isinstance(arrow_length, Angle):
            raise TypeError("arrow_length must be an Angle or None")
    if scale_length is not None:
        if not isinstance(scale_length, Angle):
            raise TypeError("scale_length must be an Angle")
    if not isinstance(wcs2d, WCS):
        raise TypeError("wcs2d must be a WCS object.")
    if scale_location not in [0, 1, 2, 3, 4]:
        raise ValueError(f"Invalid scale_location value {scale_location}.")
    if parangle is not None:
        if not isinstance(parangle, Angle):
            raise TypeError("parangle must be an Angle")

    if verbose:
        print(wcs2d)

    pixel_scales = proj_plane_pixel_scales(wcs2d)
    if not np.isclose(pixel_scales[0], pixel_scales[1]):
        raise ValueError(f'{pixel_scales[0]=} must be equal to {pixel_scales[1]=} to draw NE arrows')
    pixel_scale = pixel_scales[0] * Unit('deg')

    # origin of the NE arrows
    x_center, y_center = wcs2d.wcs.crpix
    center_coord = wcs2d.pixel_to_world(x_center - 1, y_center - 1)
    if verbose:
        print(f'{x_center=}, {y_center=} (FITS criterion)')
        print(f'{center_coord=}')

    # current plot limits
    xmin, xmax = ax.get_xlim()
    dx = xmax - xmin
    ymin, ymax = ax.get_ylim()
    dy = ymax - ymin
    diagonal_pixels = np.sqrt(dx * dx + dy * dy)
    diagonal_deg = diagonal_pixels * pixel_scale
    # define relevant lengths if necessary
    if arrow_length is None:
        arrow_length = Angle(diagonal_pixels / 8 * pixel_scale)
    if scale_length is None:
        scale_length = arrow_length
    parangle_length = Angle(diagonal_pixels / 1.5 * pixel_scale)   # 1.5 < sqrt(2)=1.41


    # define North and East direction
    tgap = 1.10
    north = SkyCoord(ra=center_coord.ra, dec=center_coord.dec + arrow_length)
    north_text = SkyCoord(ra=center_coord.ra, dec=center_coord.dec + tgap * arrow_length)
    east = SkyCoord(ra=center_coord.ra + arrow_length/np.cos(center_coord.dec), dec=center_coord.dec)
    east_text = SkyCoord(ra=center_coord.ra + tgap * arrow_length/np.cos(center_coord.dec), dec=center_coord.dec)

    # convert celestial coordinates to pixel coordinates
    north_pix = wcs2d.world_to_pixel(north)
    north_text_pix = wcs2d.world_to_pixel(north_text)
    east_pix = wcs2d.world_to_pixel(east)
    east_text_pix = wcs2d.world_to_pixel(east_text)

    # FITS criterion
    north_pix = (north_pix[0] + 1, north_pix[1] + 1)
    north_text_pix = (north_text_pix[0] + 1, north_text_pix[1] + 1)
    east_pix = (east_pix[0] + 1, east_pix[1] + 1)
    east_text_pix = (east_text_pix[0] + 1, east_text_pix[1] + 1)
    if verbose:
        print(f'North pixel.....: {north_pix} (FITS criterion)')
        print(f'North text pixel: {north_text_pix} (FITS criterion)')
        print(f'East  pixel: {east_pix} (FITS criterion)')
        print(f'East  text pixel: {east_text_pix} (FITS criterion)')

    # North and East vectors
    if fits_criterion:
        xyoffset = 0
    else:
        xyoffset = 1
    ax.plot(np.array([x_center - xyoffset, north_pix[0] - xyoffset]),
            np.array([y_center - xyoffset, north_pix[1] - xyoffset]), '-', color=color)
    ax.text(north_text_pix[0] - xyoffset, north_text_pix[1] - xyoffset,
            'N', color=color, ha='center', va='center')
    ax.plot(np.array([x_center - xyoffset, east_pix[0] - xyoffset]),
            np.array([y_center - xyoffset, east_pix[1] - xyoffset]), '-', color=color)
    ax.text(east_text_pix[0] - xyoffset, east_text_pix[1] - xyoffset,
            'E', color=color, ha='center', va='center')

    # parallactic angle
    if parangle is not None:
        for theta in [0, 180]:
            new_north = center_coord.directional_offset_by(
                position_angle=parangle + theta * Unit('degree'),
                separation=parangle_length / 2
            )
            new_north_pix = wcs2d.world_to_pixel(new_north)
            new_north_pix = (new_north_pix[0] + 1, new_north_pix[1] + 1)
            ax.plot(np.array([x_center - xyoffset, new_north_pix[0] - xyoffset]),
                    np.array([y_center - xyoffset, new_north_pix[1] - xyoffset]),
                    linestyle='--', color=color_parangle)

    # scale
    if scale_location != 0:
        xmin, xmax = ax.get_xlim()
        dx = xmax - xmin
        ymin, ymax = ax.get_ylim()
        dy = ymax - ymin
        scale_pixels = scale_length / pixel_scale
        if scale_location in [1, 4]:
            x_scale_max = xmax - dx / 20
            x_scale_min = x_scale_max - scale_pixels
        else:
            x_scale_min = xmin + dx / 20
            x_scale_max = x_scale_min + scale_pixels
        if scale_location in [1, 2]:
            y_scale = ymax - dy / 10
        else:
            y_scale = ymin + dy / 20

        if color_scale is None:
            color_scale = color
        ax.plot(np.array([x_scale_min, x_scale_max]),
                np.array([y_scale, y_scale]),
                color=color_scale)
        ax.plot(np.array([x_scale_min, x_scale_min]),
                np.array([y_scale-dy/80, y_scale+dy/80]),
                color=color_scale)
        ax.plot(np.array([x_scale_max, x_scale_max]),
                np.array([y_scale-dy/80, y_scale+dy/80]),
                color=color_scale)
        scale_font_size = ax.xaxis.get_label().get_fontsize() * fontsize_scale_relative_factor
        if diagonal_deg.value * 3600 < 60:
            label = f"{scale_length.to(Unit('arcsec')):.2f}"
        elif diagonal_deg.value * 60 < 60:
            label = f"{scale_length.to(Unit('arcmin')):.2f}"
        else:
            label = f"{scale_length.to(Unit('degree')):.2f}"
        ax.text((x_scale_min + x_scale_max) / 2, y_scale + 0.02*dy, label,
                fontsize=scale_font_size, ha='center', va='bottom', color=color_scale)





import numpy
import astropy.io.fits as fits
import numina.processing as proc


def test_bias_corrector():
    biasmap = numpy.zeros((10, 10)) + 200.0
    calibid = 'testbiasmapid'
    corrector = proc.BiasCorrector(biasmap=biasmap, calibid=calibid)

    data = numpy.zeros((10, 10)) + 600
    result = numpy.zeros((10, 10)) + 400
    hdu = fits.PrimaryHDU(data)
    hdu.header['DATE-OBS'] = '2010-03-11T04:34:12'
    hdul = fits.HDUList([hdu])
    newhdul = corrector.run(hdul)

    assert newhdul[0].header['NUM-BS'] == calibid
    assert numpy.allclose(newhdul[0].data, result)


def test_dark_corrector():
    darkmap = numpy.zeros((10, 10)) + 2.0
    calibid = 'testdarkmapid'
    corrector = proc.DarkCorrector(darkmap=darkmap, calibid=calibid)

    data = numpy.zeros((10, 10)) + 102.0
    result = numpy.zeros((10, 10)) + 100.0
    hdu = fits.PrimaryHDU(data)
    hdu.header['DATE-OBS'] = '2010-03-11T04:34:12'
    hdul = fits.HDUList([hdu])
    newhdul = corrector.run(hdul)

    assert newhdul[0].header['NUM-DK'] == calibid
    assert numpy.allclose(newhdul[0].data, result)



#
# Copyright 2010-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import logging

import numina.util.node as node


_logger = logging.getLogger(__name__)


class FlowError(Exception):
    """Error base class for flows."""
    pass


class SerialFlow(node.Node):
    """A flow where Nodes are executed sequentially."""

    def __init__(self, nodeseq):
        # Checking inputs and out puts are correct
        self.nodeseq = nodeseq
        if nodeseq:
            ninputs = nodeseq[0].ninputs
            noutputs = nodeseq[-1].noutputs
            # Check nodes
            for i, o in zip(nodeseq, nodeseq[1:]):
                if i.noutputs != o.ninputs:
                    raise FlowError
        else:
            ninputs = 1
            noutputs = 1
        super(SerialFlow, self).__init__(ninputs, noutputs)

    def __iter__(self):
        return self.nodeseq.__iter__()

    def __len__(self):
        return self.nodeseq.__len__()

    def __getitem__(self, key):
        return self.nodeseq[key]

    def __setitem__(self, key, value):
        self.nodeseq[key] = value

    def run(self, img):
        out = img
        for nd in self.nodeseq:
            out = nd(out)
        return out


class ParallelFlow(node.Node):
    """A flow where Nodes are executed in parallel."""

    def __init__(self, nodeseq):
        self.nodeseq = nodeseq
        nin = sum((f.ninputs for f in nodeseq), 0)
        nout = sum((f.noutputs for f in nodeseq), 0)
        super(ParallelFlow, self).__init__(nin, nout)

    def run(self, args):
        out = []
        for func, arg in zip(self.nodeseq, args):
            r = func(arg)
            out.append(r)

        if any(f.noutputs != 1 for f in self.nodeseq):
            return tuple(item for sublist in out for item in sublist)

        return tuple(out)

    def __iter__(self):
        return self.nodeseq.__iter__()

    def __len__(self):
        return self.nodeseq.__len__()

    def __getitem__(self, key):
        return self.nodeseq[key]

    def __setitem__(self, key, value):
        self.nodeseq[key] = value


class MixerFlow(node.Node):
    def __init__(self, table):
        nin = max(table) + 1
        nout = len(table)
        super(MixerFlow, self).__init__(nin, nout)
        self.table = table

    def run(self, args):
        assert len(args) == self.ninputs

        return tuple(args[idx] for idx in self.table)




import numina.core.tagexpr as tagexpr


def test_tag_expresion1():

    mode_tags = {'vph': 'MR-I', 'insmode': 'LCB'}

    insmode = tagexpr.TagRepr("insmode")
    vph = tagexpr.TagRepr("vph")
    speclamp = tagexpr.TagRepr("speclamp")
    p_ = tagexpr.Placeholder

    expr1 = vph == p_("vph")
    expr2 = (vph == "LR-I") & (insmode == "LCB") & (insmode != "MOS")
    expr3 = (vph == "LR-I") & (insmode == "LCB") & (speclamp == p_("speclamp"))

    assert isinstance(expr1, tagexpr.Expression)
    assert isinstance(expr2, tagexpr.Expression)
    assert isinstance(expr3, tagexpr.Expression)

    expr11 = expr1.fill_placeholders(**mode_tags)
    assert len(expr11.places()) == 0

    expr21 = expr2.fill_placeholders(**mode_tags)
    assert len(expr21.places()) == 0

    # with pytest.raises(KeyError):
    #    expr3.fill_placeholders(**mode_tags)


def test_tag_expresion_filter1():

    insmode = tagexpr.TagRepr("insmode")
    vph = tagexpr.TagRepr("vph")
    p_ = tagexpr.Placeholder

    expr1 = (vph == "LR-I") & (insmode == p_("insmode")) & (insmode != "MOS")

    res = [n for n in tagexpr.filter_tree(lambda node: True, expr1)]

    assert len(res) == 11

    assert isinstance(res[0], tagexpr.TagRepr)
    assert res[0].name == "vph"
    assert isinstance(res[1], tagexpr.ConstExpr)
    assert res[1].value == "LR-I"
    assert isinstance(res[2], tagexpr.PredEq)
    assert isinstance(res[3], tagexpr.TagRepr)
    assert res[3].name == "insmode"
    assert isinstance(res[4], tagexpr.Placeholder)
    assert res[4].name == "insmode"
    assert isinstance(res[5], tagexpr.PredEq)
    assert isinstance(res[6], tagexpr.PredAnd)
    assert isinstance(res[7], tagexpr.TagRepr)
    assert res[7].name == "insmode"
    assert isinstance(res[8], tagexpr.ConstExpr)
    assert res[8].value == "MOS"
    assert isinstance(res[9], tagexpr.PredNe)
    assert isinstance(res[10], tagexpr.PredAnd)

    expr2 = expr1.fill_placeholders(insmode='LCB')
    assert expr2.eval(insmode='LCB', vph='LR-I') is True


def test_tag_expresion_filter2():

    insmode = tagexpr.TagRepr("insmode")
    vph = tagexpr.TagRepr("vph")
    temp = tagexpr.TagRepr("temp")
    p_ = tagexpr.Placeholder

    expr1 = (vph == p_("vph")) & (
        insmode == p_("insmode")) & (temp >= p_("temp"))
    expr2 = expr1.fill_placeholders(insmode='LCB', vph='LR-I', temp=22.4)

    assert expr2.eval(insmode='LCB', vph='LR-I', temp=22.4)

    assert expr2.eval(insmode='LCB', vph='LR-I', temp=24.4)

    assert expr2.eval(insmode='LCB', vph='LR-I', temp=20.0) is False

    assert expr2.eval(insmode='MOS', vph='LR-I', temp=23.0) is False



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

ctypedef fused datacube_t:
    double[:,:,:]
    float[:,:,:]
    long[:,:,:]
    int[:,:,:]

ctypedef fused result_t:
    double[:,:]
    float[:,:]

ctypedef char[:,:] mask_t

MASK_GOOD = 0
MASK_SATURATION = 3




#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""User command line interface of Numina."""

import warnings

import numina
import numina.drps
from numina.user.clishowins import print_no_instrument


def register(subparsers, config):
    parser_show_rec = subparsers.add_parser(
        'show-recipes',
        help='show information of recipes'
    )

    parser_show_rec.set_defaults(command=show_recipes, template=False)

    parser_show_rec.add_argument(
        '-i', '--instrument',
        help='filter recipes by instrument'
    )
    parser_show_rec.add_argument(
        '-m', '--mode',
        help='filter recipes by mode name'
    )
#    parser_show_rec.add_argument('--output', type=argparse.FileType('wb', 0))

    parser_show_rec.add_argument(
        'name', nargs='*', default=None,
        help='filter recipes by name'
    )

    return parser_show_rec


def show_recipes(args, extra_args, config):

    drpsys = numina.drps.get_system_drps()

    # Query instruments
    if args.instrument:
        name = args.instrument
        try:
            val = drpsys.query_by_name(name)
        except KeyError:
            val = None
        res = [(name, val)]
    else:
        res = drpsys.query_all().items()

    # predicates
    preds = []
    if args.name:
        def pred1(mode_rec): return mode_rec[1]['class'] in args.name  # noqa: E731
        preds.append(pred1)
    if args.mode:
        def pred1(mode_rec): return mode_rec[0] == args.mode  # noqa: E731
        preds.append(pred1)

    for name, theins in res:
        # Per instrument
        if theins:
            for pipe in theins.pipelines.values():
                for mode, recipe_entry in pipe.recipes.items():
                    mod_rec = mode, recipe_entry
                    # Check all predicates
                    for pre in preds:
                        if not pre(mod_rec):
                            break
                    else:
                        recipe_fqn = recipe_entry['class']
                        recipe = pipe.get_recipe_object(mode)
                        print_recipe(
                            recipe.__class__, name=recipe_fqn,
                            insname=theins.name,
                            pipename=pipe.name,
                            modename=mode
                        )
        else:
            print_no_instrument(name)


def print_requirements(recipe, pad=''):

    for req in recipe.requirements().values():
        if req.hidden:
            # I Do not want to print it
            continue
        dispname = req.dest

        if req.optional:
            dispname = dispname + '(optional)'

        if req.default is not None:
            dispname = dispname + '=' + str(req.default)
        typ = req.type.descriptive_name()

        print(f"{pad}{dispname} type={typ!r} [{req.description}]")


def print_recipe(recipe, name=None, insname=None,
                 pipename=None, modename=None):
    try:
        if name is None:
            name = recipe.__module__ + '.' + recipe.__name__
        print('Recipe:', name)
        if recipe.__doc__:
            print(' summary:',
                  recipe.__doc__.lstrip().expandtabs().splitlines()[0]
                  )
        if insname:
            print(' instrument:', insname)
        if pipename:
            print('  pipeline:', pipename)
        if modename:
            print('  obs mode:', modename)
        print(' requirements:')
        print_requirements(recipe, pad='  ')
        print()
    except Exception as error:
        warnings.warn(f'problem {error} with recipe {recipe!r}')



#
# Copyright 2017-2023 Universidad Complutense de Madrid
#
# This file is part of Numina DRP
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Unit test for MultiType adapter"""

import pytest

import numina.exceptions
from ..datatype import DataType
from ..multitype import MultiType
from ..structured import BaseStructuredCalibration


class AVal(object):
    pass


class BVal(object):
    pass


class TypeA(DataType):

    def __init__(self):
        super(TypeA, self).__init__(ptype=AVal)

    def convert(self, obj):
        return TypeA()

    def tag_names(self):
        return ['domeA', 'domeA2']


class TypeB(DataType):

    def __init__(self):
        super(TypeB, self).__init__(ptype=BVal)

    def convert(self, obj):
        return TypeB()

    def tag_names(self):
        return ['domeB']


class Structured1(BaseStructuredCalibration):
    def __init__(self):
        super(Structured1, self).__init__()
        self.struct1 = 1

    def __getstate__(self):
        state = super(Structured1, self).__getstate__()
        state['struct1'] = self.struct1
        return state

    def __setstate__(self, state):
        super(Structured1, self).__setstate__(state)
        self.struct1 = state['struct1']


class Structured2(BaseStructuredCalibration):
    def __init__(self):
        super(Structured2, self).__init__()
        self.struct2 = 2

    def __getstate__(self):
        state = super(Structured2, self).__getstate__()
        state['struct2'] = self.struct2
        return state

    def __setstate__(self, state):
        super(Structured2, self).__setstate__(state)
        self.struct2 = state['struct2']


class Structured3(BaseStructuredCalibration):
    def __init__(self):
        super(Structured3, self).__init__()
        self.struct3 = 3

    def __getstate__(self):
        state = super(Structured3, self).__getstate__()
        state['struct3'] = self.struct3
        return state

    def __setstate__(self, state):
        super(Structured3, self).__setstate__(state)
        self.struct3 = state['struct3']


# def test_convert_in():
#
#     multi = MultiType(TypeA, TypeB)
#
#     with pytest.raises(ValueError):
#         multi.convert_in(None)


def test_validate1():
    multi = MultiType(TypeA, TypeB)

    # testing without selection the internal type
    # It should validate both types
    assert multi.validate(AVal())
    assert multi.validate(BVal())

    # And fail with anything else
    with pytest.raises(numina.exceptions.ValidationError) as excinfo:
        multi.validate(3)

    intl_exception = excinfo.value
    intl_args = intl_exception.args[1]
    assert intl_args == (AVal, BVal)


def test_validate2():

    multi = MultiType(TypeA, TypeB)

    # Simulate valid query
    multi.current_node = multi.node_type[0]

    # testing with selection of the internal type
    # It should validate TypeA only
    assert multi.validate(AVal())

    with pytest.raises(numina.exceptions.ValidationError):
        multi.validate(BVal())

    with pytest.raises(numina.exceptions.ValidationError):
        multi.validate(3)


def test_tag_names1():

    multi = MultiType(TypeA, TypeB)

    assert multi.tag_names() == ['domeA', 'domeA2', 'domeB']


def test_tag_names2():

    multi = MultiType(TypeA, TypeB)
    multi.current_node = multi.node_type[1]

    assert multi.tag_names() == ['domeB']


def test_internal_default():
    import numina.core.dataholders as dh

    multi = MultiType(TypeA, TypeB)
    req = dh.Requirement(multi, description="Some")

    with pytest.raises(ValueError):
        req.default_value()


def test_internal_default_optional():
    import numina.core.dataholders as dh

    multi = MultiType(TypeA, TypeB)
    req = dh.Requirement(multi, description="Some", optional=True)

    assert req.default_value() is None


def test_load_multitype(tmpdir):
    import numina.store
    import numina.util.context as cntx

    multi = MultiType(Structured1, Structured2)

    prefix1 = 'some1'
    prefix2 = 'some2'
    prefix3 = 'some3'

    filename1 = prefix1 + '.json'
    filename2 = prefix2 + '.json'
    filename3 = prefix3 + '.json'

    obj1 = Structured1()
    obj2 = Structured2()
    obj3 = Structured3()

    with cntx.working_directory(str(tmpdir)):
        numina.store.dump(obj1, obj1, prefix1)
        numina.store.dump(obj2, obj2, prefix2)
        numina.store.dump(obj3, obj3, prefix3)

    with cntx.working_directory(str(tmpdir)):

        recover = numina.store.load(multi, filename1)
        assert isinstance(recover, Structured1)

        recover = numina.store.load(multi, filename2)
        assert isinstance(recover, Structured2)

        with pytest.raises(TypeError):
            numina.store.load(multi, filename3)



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
from astropy.units import Unit
import numpy as np
import os

from .raise_valueerror import raise_ValueError


def load_atmosphere_transmission_curve(atmosphere_transmission, wmin, wmax, wv_cunit1, faux_dict, verbose):
    """Load atmosphere transmission curve.

    Parameters
    ----------
    atmosphere_transmission : str
        String indicating whether the atmosphere transmission of
        the atmosphere is applied or not. Two possible values are:
        - 'default': use default curve defined in 'faux_dict'
        - 'none': do not apply atmosphere transmission
    wmin : `~astropy.units.Quantity`
        Minimum wavelength to be considered.
    wmax : `~astropy.units.Quantity`
        Maximum wavelength to be considered.
    wv_cunit1 : `~astropy.units.core.Unit`
        Default wavelength unit to be employed in the wavelength scale.
    faux_dict : Python dictionary
        File names of auxiliary files:
        - skycalc: table with SKYCALC Sky Model Calculator predictions
        - flatpix2pix: pixel-to-pixel flat field
        - model_ifu2detector: 2D polynomial transformation
          x_ifu, y_ify, wavelength -> x_detector, y_detector
    verbose : bool
        If True, display additional information.

    Returns
    -------
    wave_transmission : `~astropy.units.Quantity`
        Wavelength column of the tabulated transmission curve.
    curve_transmission : `~astropy.units.Quantity`
        Transmission values for the wavelengths given in
        'wave_transmission'.

    """

    if atmosphere_transmission == "default":
        infile = faux_dict['skycalc']
        if verbose:
            print(f'\nLoading atmosphere transmission curve {os.path.basename(infile)}')
        with fits.open(infile) as hdul:
            skycalc_header = hdul[1].header
            skycalc_table = hdul[1].data
        if skycalc_header['TTYPE1'] != 'lam':
            raise_ValueError(f"Unexpected TTYPE1: {skycalc_header['TTYPE1']}")
        cwave_unit = skycalc_header['TUNIT1']
        wave_transmission = skycalc_table['lam'] * Unit(cwave_unit)
        curve_transmission = skycalc_table['trans']
        if wmin < np.min(wave_transmission) or wmax > np.max(wave_transmission):
            print(f'{wmin=} (simulated photons)')
            print(f'{wmax=} (simulated photons)')
            print(f'{np.min(wave_transmission.to(wv_cunit1))=} (transmission curve)')
            print(f'{np.max(wave_transmission.to(wv_cunit1))=} (transmission curve)')
            raise_ValueError('Wavelength range covered by the tabulated transmission curve is insufficient')
    elif atmosphere_transmission == "none":
        wave_transmission = None
        curve_transmission = None
        if verbose:
            print('Skipping application of the atmosphere transmission')
    else:
        wave_transmission = None   # avoid PyCharm warning (not aware of raise ValueError)
        curve_transmission = None  # avoid PyCharm warning (not aware of raise ValueError)
        raise_ValueError(f'Unexpected {atmosphere_transmission=}')

    return wave_transmission, curve_transmission



"""Coadd several FITS images or subsets of FITS images listed in a TXT file"""

import argparse

from astropy.io import fits
import numpy as np

from numina.array.display.fileinfo import list_fileinfo_from_txt
from .subsets_of_fileinfo_from_txt import subsets_of_fileinfo_from_txt


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()
    parser.add_argument("input_list",
                        help="txt file with list of images to be coadded")
    # if output_fits_filename == "@" the script expects a line in
    # the file input_list of the form:
    # @ <number> <label>
    # where <number> is the number of images to be coadded and
    # <label> is the name of the output file
    parser.add_argument("output_fits_filename",
                        help="filename of output FITS image, or @ symbol")
    parser.add_argument("--factorcol",
                        help="Column number with multiplicative factors",
                        type=int)
    parser.add_argument("--offsetcol",
                        help="Column number with offsets:\n "
                             "final scan = old scan + offset (default=None)",
                        type=int)
    parser.add_argument("--average",
                        help="average result",
                        action="store_true")
    parser.add_argument("--add_header",
                        help="add header of first image: yes/no (default=yes)",
                        default="yes")
    parser.add_argument("--noclobber",
                        help="avoid overwriting existing file",
                        action='store_true')
    parser.add_argument("--out_nsum",
                        help="filename for output nsum FITS image "
                             "(default=None)",
                        type=argparse.FileType('w'),
                        default=None)
    parser.add_argument("--debugplot",
                        help="integer indicating plotting/debugging" +
                        " (default=10)",
                        type=int, default=10,
                        choices=[0, 1, 2, 10, 11, 12, 21, 22])
    args = parser.parse_args(args)

    debugplot = int(args.debugplot)

    if args.output_fits_filename == "@":
        dict_of_fileinfo = subsets_of_fileinfo_from_txt(args.input_list)
    else:
        list_of_fileinfo = list_fileinfo_from_txt(args.input_list)
        dict_of_fileinfo = {}
        dict_of_fileinfo[0] = {
            'label': args.output_fits_filename,
            'list_of_fileinfo': list_of_fileinfo
        }

    number_of_output_files = len(dict_of_fileinfo)

    # main loop in number of output files
    for iout in range(number_of_output_files):
        # list of files to be combined
        tmpdict = dict_of_fileinfo[iout]
        output_fits_filename = tmpdict['label']
        list_of_files = [tmp.filename for tmp in tmpdict['list_of_fileinfo']]
        list_of_infos = [tmp.fileinfo for tmp in tmpdict['list_of_fileinfo']]

        # number of files to be combined
        number_of_files = len(list_of_files)

        # if there are offsets between the images, check that all
        # the numbers are provided
        offsets = np.zeros(number_of_files, dtype=int)
        if args.offsetcol is not None:
            for i in range(number_of_files):
                offsets[i] = list_of_infos[i][args.offsetcol - 2]

        # if there are multiplicative factors, check that all
        # the numbers are provided
        multfactors = np.ones(number_of_files, dtype=float)
        if args.factorcol is not None:
            for i in range(number_of_files):
                multfactors[i] = list_of_infos[i][args.factorcol - 2]

        # declare auxiliary arrays to store image basic parameters
        naxis1 = np.zeros(number_of_files, dtype=int)
        naxis2 = np.zeros(number_of_files, dtype=int)

        # read basic parameters for all the images
        for i in range(number_of_files):
            infile = list_of_files[i]
            hdulist = fits.open(infile)
            image_header = hdulist[0].header
            hdulist.close()
            naxis1[i] = image_header['naxis1']
            naxis2[i] = image_header['naxis2']

        # check that NAXIS1 is the same for all the images
        naxis1_comp = np.repeat(naxis1[0], number_of_files)
        if not np.allclose(naxis1, naxis1_comp):
            raise ValueError("NAXIS1 values are different")

        # check that NAXIS2 is the same for all the images
        naxis2_comp = np.repeat(naxis2[0], number_of_files)
        if not np.allclose(naxis2, naxis2_comp):
            raise ValueError("NAXIS2 values are different")

        # declare output arrays
        image2d = np.zeros((naxis2[0], naxis1[0]))
        image2d_nsum = np.zeros((naxis2[0], naxis1[0]), dtype=int)

        image_header_first_frame = None    # avoid PyCharm warning

        # add all the individual images
        for i in range(number_of_files):
            infile = list_of_files[i]
            if abs(debugplot) >= 10:
                print("<--" + infile +
                      " (image " + str(i + 1) + " of " +
                      str(number_of_files) + ")  offset: " +
                      str(offsets[i]) + "  factor: " +
                      str(multfactors[i])
                      )
            hdulist = fits.open(infile)
            data = hdulist[0].data * multfactors[i]
            if i == 0:
                image_header_first_frame = hdulist[0].header
            hdulist.close()
            # determine the image region to be coadded taking into
            # account the corresponding offset for the i-th image
            nsmin_in = 1 + offsets[i]
            nsmax_in = naxis2[0] + offsets[i]
            if nsmin_in < 1:
                nsmin_in = 1
            if nsmax_in > naxis2[0]:
                nsmax_in = naxis2[0]
            nsmin_out = 1 - offsets[i]
            nsmax_out = naxis2[0] - offsets[i]
            if nsmin_out < 1:
                nsmin_out = 1
            if nsmax_out > naxis2[0]:
                nsmax_out = naxis2[0]
            image2d[(nsmin_in-1):nsmax_in, ] += data[(nsmin_out-1):nsmax_out, ]
            image2d_nsum[(nsmin_in-1):nsmax_in, ] += 1

        # average result when requested
        if args.average:
            image2d /= image2d_nsum
            print("==> Computing average of " + str(number_of_files) +
                  " images")
        else:
            print("==> Coadding " + str(number_of_files) +
                  " images")

        # save results
        if abs(debugplot) >= 10:
            print("==> Generating output file: " +
                  output_fits_filename + "...")
        if args.add_header == "yes":
            hdu = fits.PrimaryHDU(image2d, image_header_first_frame)
        else:
            hdu = fits.PrimaryHDU(image2d)
        hdu.writeto(output_fits_filename, overwrite=(not args.noclobber))

        if args.out_nsum is not None:
            if abs(debugplot) >= 10:
                print("==> Generating output file: " +
                      args.out_nsum.name + "...")
            hdu = fits.PrimaryHDU(image2d_nsum.astype(np.int16))
            hdu.writeto(args.out_nsum, overwrite=(not args.noclobber))

        if abs(debugplot) >= 10:
            if args.average == "yes":
                print("average OK!")
            else:
                print("sum OK!")


if __name__ == "__main__":

    main()



import contextlib
import os


@contextlib.contextmanager
def working_directory(path):
    """A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.

    """

    # http://code.activestate.com/recipes/576620-changedirectory-context-manager/

    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


@contextlib.contextmanager
def ignored(*exceptions):

    # by Raymond Hettinger
    # https://www.youtube.com/watch?v=OSGv2VnC0go&t=1215s

    try:
        yield
    except exceptions:
        pass


@contextlib.contextmanager
def environ(**keys):
    """Context manager that records environment"""
    old_vals = {key: os.environ.get(key, None) for key in keys}
    os.environ.update(keys)
    try:
        yield
    finally:
        for key, val in old_vals.items():
            if val is None:
                del os.environ[key]
            else:
                os.environ[key] = val


@contextlib.contextmanager
def manage_fits(list_of_frame):
    """Manage a list of FITS resources"""

    import astropy.io.fits as fits
    import numina.types.dataframe as df

    refs = []
    for frame in list_of_frame:
        if isinstance(frame, str):
            ref = fits.open(frame)
            refs.append(ref)
        elif isinstance(frame, fits.HDUList):
            refs.append(frame)
        elif isinstance(frame, df.DataFrame):
            ref = frame.open()
            refs.append(ref)
        else:
            refs.append(frame)
    try:
        yield refs
    finally:
        # release
        for obj in refs:
            obj.close()



#
# Copyright 2015-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Unit test for taggers"""

import pytest

import astropy.io.fits as fits

from numina.datamodel import DataModel
from numina.types.dataframe import DataFrame
from ..oresult import ObservationResult
from ..taggers import extract_tags_from_obsres


@pytest.fixture
def my_datamodel():
    mappings = {'filter': 'filter', 'read_mode': 'readm'}
    model = DataModel('TEST', mappings=mappings)
    return model


def test_empty_ob(my_datamodel):

    ob = ObservationResult()
    tags = extract_tags_from_obsres(ob, tag_keys=[], datamodel=my_datamodel)

    assert len(tags) == 0


def test_init_ob(my_datamodel):

    img1 = fits.PrimaryHDU(data=[1, 2, 3])
    frame1 = DataFrame(frame=fits.HDUList(img1))

    ob = ObservationResult()
    ob.frames = [frame1]
    tags = extract_tags_from_obsres(ob, tag_keys=[], datamodel=my_datamodel)

    assert len(tags) == 0


def test_header_key1_ob(my_datamodel):

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD1'
    frame1 = DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD2'
    frame2 = DataFrame(frame=fits.HDUList(img1))

    ob = ObservationResult()
    ob.frames = [frame1, frame2]
    tags = extract_tags_from_obsres(ob, tag_keys=['filter'], datamodel=my_datamodel)

    assert tags == {'filter': 'FILTER-A'}


def test_header_key1_mis(my_datamodel):
    """Test extract_tags raises ValueError if there is a missmatch"""
    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    frame1 = DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['FILTER'] = 'FILTER-B'
    frame2 = DataFrame(frame=fits.HDUList(img2))

    ob = ObservationResult()
    ob.frames = [frame1, frame2]

    with pytest.raises(ValueError):
        extract_tags_from_obsres(ob, tag_keys=['filter'], datamodel=my_datamodel)


def test_header_key1_mis_no_strict(my_datamodel):
    """Test extract_tags returns is struct is false"""
    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    frame1 = DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['FILTER'] = 'FILTER-B'
    frame2 = DataFrame(frame=fits.HDUList(img2))

    ob = ObservationResult()
    ob.frames = [frame1, frame2]

    tags = extract_tags_from_obsres(ob, tag_keys=['filter'], datamodel=my_datamodel, strict=False)
    assert tags == {'filter': 'FILTER-A'}


def test_header_key2_ob(my_datamodel):

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD1'
    frame1 = DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD1'
    frame2 = DataFrame(frame=fits.HDUList(img1))

    ob = ObservationResult()
    ob.frames = [frame1, frame2]
    tags = extract_tags_from_obsres(ob, tag_keys=['filter', 'read_mode'], datamodel=my_datamodel)

    assert tags == {'filter': 'FILTER-A', 'read_mode': 'MOD1'}



#
# Copyright 2016-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


class PropertyBase:
    def __init__(self, is_readonly=True, mapping=None, depends=None):
        self.is_readonly = is_readonly
        self.mapping = mapping
        self.name = "undefined"
        #
        if depends is None:
            self.depends = []
        else:
            self.depends = depends

    def __set_name__(self, owner, name):
        self.name = name


class PropertyEntry(PropertyBase):
    """Class representing a property entry"""

    def __init__(self, values, depends):
        super().__init__(depends=depends)
        self.values = values

    def get(self, **kwds):
        result = self.values
        for dep in self.depends:
            key = kwds[dep]
            result = result[key]
        return result

    def __get__(self, instance, owner=None):
        state = {dep: None for dep in self.depends}
        obj = instance
        while obj is not None:
            early_exit = True
            for dep in self.depends:
                if state[dep] is None:
                    if hasattr(obj, dep):
                        state[dep] = getattr(obj, dep)
                    else:
                        early_exit = False
            if early_exit:
                # we have filled all Nones
                break
            # Find top level
            obj = obj.parent

        result = self.values
        for dep in self.depends:
            key = state[dep]
            result = result[key]
        return result

    def __set__(self, obj, value):
        pass

    def defaults(self):
        val = {}
        result = self.values
        for dep in self.depends:
            # First key, sorted needed in Python < 3.6
            key = sorted(result.keys())[0]
            val[dep] = key
            result = result[key]
        return val


class PropertyProxy(PropertyBase):
    """Class representing a property proxy"""

    def __init__(self, ref_name, is_readonly=True):
        # FIXME: depends should be the depends of the ref_name
        super().__init__(is_readonly=is_readonly, mapping=None, depends=None)
        self.ref_name = ref_name

    def defaults(self):
        # FIXME: defaults should be the defaults of the ref_name
        val = {}
        return val

    def __get__(self, instance, owner=None):
        res = instance.get_property(self.ref_name)
        return res

    def __set__(self, obj, value):
        dev_path, prop_name = obj.split_path(self.ref_name)
        device = obj.get_device_seq(dev_path)
        setattr(device, prop_name, value)


class _PropertyModBase(PropertyBase):
    """Class representing a property"""

    def __init__(self, default=0, mapping=None):
        super().__init__(is_readonly=False, mapping=mapping, depends=None)
        self.default = default

    def defaults(self):
        # FIXME
        val = {}
        return val

    def value_check(self, value):  # pragma: no cover
        return True

    def __get__(self, instance, owner=None):
        intl = instance._internal_state  # noqa
        if self.name not in intl:
            intl[self.name] = self.default
        return intl[self.name]

    def __set__(self, obj, value):
        if self.value_check(value):
            obj._internal_state[self.name] = value  # noqa
        else:
            raise ValueError(f"value {value} is not allowed")


class PropertyModOneOf(_PropertyModBase):
    """Class representing a property"""

    def __init__(self, one_of, default=0, mapping=None):
        self.one_of = one_of
        super().__init__(default, mapping)

    def value_check(self, value):
        if value in self.one_of:
            return True
        else:
            return False


class PropertyModLimits(_PropertyModBase):
    """Class representing a property in a range"""

    def __init__(self, limits, default=0, mapping=None):
        match limits:
            case []:
                self.limit_1 = None
                self.limit_2 = None
            case [a, b]:
                self.limit_1 = a
                self.limit_2 = b
            case _:
                raise ValueError(f"invalid range {limits}")

        super().__init__(default, mapping)

    def value_check(self, value):
        if self.limit_1 is not None and value < self.limit_1:
            return False
        if self.limit_2 is not None and value > self.limit_2:
            return False

        return True



#
# Copyright 2018-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import logging


def logging_from_debugplot(debugplot):
    """Set debugging level based on debugplot value.

    Parameters
    ----------
    debugplot : int
            Debugging level for messages and plots. For details see
            'numina.array.display.pause_debugplot.py'.
    """

    if isinstance(debugplot, int):
        if abs(debugplot) >= 10:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)
    else:
        raise ValueError("Unexpected debugplot=" + str(debugplot))



#
# Copyright 2015-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Function to retrieve tags from Observation results."""


import itertools
import warnings

from numina.datamodel import DataModel


def get_tags_from_full_ob(ob, reqtags=None):
    """
    Parameters
    ----------
    ob (ObservationResult): Observation result
    reqtags (iterable): Keywords

    Returns
    -------
    A dictionary

    """
    warnings.warn("'get_tags_from_full_ob' is deprecated, use 'extract_tags_from_obsres' instead",
                  DeprecationWarning, stacklevel=2)


    if reqtags is None:
        reqtags = []
    # adding an override from reqtags
    mappings = {r: r for r in reqtags}
    datamodel = DataModel(mappings=mappings)

    return extract_tags_from_obsres(ob, reqtags, datamodel, strict=True)


def extract_tags_from_obsres(obsres, tag_keys, datamodel: DataModel, strict=True) -> dict:
    sample = obsres.get_sample_frame()
    if sample is None:
        return {}
    ref_img = sample.open()
    final_tags = extract_tags_from_img(
        ref_img, tag_keys, datamodel, base=obsres.labels)
    if strict:
        for frame in itertools.chain(obsres.frames, obsres.results.values()):
            this_tags = extract_tags_from_img(
                frame.open(), tag_keys, datamodel, base=obsres.labels)
            if this_tags != final_tags:
                raise ValueError(f"tags in image {frame} are {this_tags} ! = {final_tags}")

        for res in obsres.children:
            res_tags = res.tags
            for t in tag_keys:
                if final_tags[t] != res_tags[t]:
                    msg = f'wrong tag {t} in product {res}'
                    raise ValueError(msg)

        return final_tags
    else:
        return final_tags


def extract_tags_from_img(img, tag_keys, datamodel: DataModel, base=None) -> dict:

    base = base or {}
    fits_extractor = datamodel.extractor_map['fits']
    final_tags = {}
    for key in tag_keys:

        if key in base:
            final_tags[key] = base[key]
        else:
            final_tags[key] = fits_extractor.extract(key, img)
    return final_tags



#
# Copyright 2008-2014 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

include "procdefs.pyx"
include "rampproc.pyx"
include "fowlerproc.pyx"




import astropy.io.fits as fits

from ..oresult import ObservationResult
import numina.datamodel
import numina.core
import numina.types.dataframe as dataframe


def test_oresult_empty():

    datamodel = numina.datamodel.DataModel()
    ob = ObservationResult()
    meta = ob.metadata_with(datamodel)
    assert meta == {'info': []}


def test_oresult1():

    datamodel = numina.datamodel.DataModel()
    ob = ObservationResult()
    for i in range(3):
        hdu = fits.PrimaryHDU()
        hdu.header['BLCKUUID'] = '1'
        hdu.header['INSTRUME'] = 'TEST'
        hdu.header['UUID'] = '1'
        hdu.header['DATE-OBS'] = '2018-04-12T22:44:12.3'
        hdu.header['INSCONF'] = 'v1'
        hdu.header['FILENAME'] = f'img{i}.fits'
        hdu.header['EXPTIME'] = 3.0

        hdulist = fits.HDUList(hdu)
        ob.frames.append(dataframe.DataFrame(frame=hdulist))

    meta = ob.metadata_with(datamodel)

    assert meta['block_uuid'] == '1'
    assert len(meta['info']) == 3



#
# Copyright 2014-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np
from astropy.modeling import Fittable1DModel, Parameter


def sum_of_gaussian_factory(N):
    """Return a model of the sum of N Gaussians and a constant background."""

    name = f"SumNGauss{N:d}"
    attr = {}

    # parameters
    for i in range(N):
        key = f"amplitude_{i:d}"
        attr[key] = Parameter(key)
        key = f"center_{i:d}"
        attr[key] = Parameter(key)
        key = f"stddev_{i:d}"
        attr[key] = Parameter(key)

    attr['background'] = Parameter('background', default=0.0)

    def fit_eval(self, x, *args):
        result = x * 0 + args[-1]
        for i in range(N):
            result += args[3 * i] * \
                np.exp(- 0.5 * (x - args[3 * i + 1])
                       ** 2 / args[3 * i + 2] ** 2)
        return result

    attr['evaluate'] = fit_eval

    def deriv(self, x, *args):
        d_result = np.ones(((3 * N + 1), len(x)))

        for i in range(N):
            d_result[3 * i] = (np.exp(-0.5 / args[3 * i + 2] ** 2 *
                                      (x - args[3 * i + 1]) ** 2))
            d_result[3 * i + 1] = args[3 * i] * d_result[3 * i] * \
                (x - args[3 * i + 1]) / args[3 * i + 2] ** 2
            d_result[3 * i + 2] = args[3 * i] * d_result[3 * i] * \
                (x - args[3 * i + 1]) ** 2 / args[3 * i + 2] ** 3
        return d_result

    attr['fit_deriv'] = deriv

    klass = type(name, (Fittable1DModel, ), attr)
    return klass
